"""Database configuration UI components.

This subpackage exposes the main entry points for database-related
configuration UIs.
"""

from .db_selection import select_db_backend
from .generic import get_config
from .main import configure_db
from .onprem import get_onprem
from .saas import get_saas

__all__ = [
    "get_config",
    "configure_db",
    "select_db_backend",
    "get_onprem",
    "get_saas",
]



from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import (
    ScsOption,
    ScsSecretOption,
)

EXTRA_SAAS_OPTIONS = [
    ScsOption(
        "--saas-url",
        metavar="URL",
        type=str,
        default="https://cloud.exasol.com",
        help="Exasol SaaS service URL",
        scs_key=CKey.saas_url,
    ),
    ScsOption(
        "--saas-account-id",
        metavar="ACCOUNT_ID",
        type=str,
        help="Exasol SaaS account ID",
        scs_key=CKey.saas_account_id,
    ),
    # CKey.saas_database_id and CKey.saas_database_name can be used
    # alternatively, see
    # https://github.com/exasol/saas-api-python/blob/main/exasol/saas/client/api_access.py#L134
    ScsOption(
        "--saas-database-id",
        metavar="ID",
        type=str,
        help="Exasol SaaS database ID",
        scs_key=CKey.saas_database_id,
        scs_alternative_key=CKey.saas_database_name,
    ),
    ScsOption(
        "--saas-database-name",
        metavar="NAME",
        type=str,
        help="Exasol SaaS database name",
        scs_key=CKey.saas_database_name,
        scs_alternative_key=CKey.saas_database_id,
    ),
    ScsSecretOption(
        "--saas-token",
        envvar="SCS_EXASOL_SAAS_TOKEN",
        prompt="Exasol SaaS personal access token (PAT)",
        scs_key=CKey.saas_token,
        metavar="PAT",
    ),
]



from importlib.metadata import version

from exasol_transformers_extension.deployment.scripts_deployer import ScriptsDeployer
from exasol_transformers_extension.deployment.te_language_container_deployer import (
    TeLanguageContainerDeployer,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.bfs_connection import ensure_bfs_connection
from exasol.nb_connector.connections import open_pyexasol_connection
from exasol.nb_connector.extension_wrapper_common import (
    PATH_IN_BUCKET_FOR_SLC,
    deploy_language_container,
    encapsulate_huggingface_token,
)
from exasol.nb_connector.language_container_activation import (
    ACTIVATION_KEY_PREFIX,
    get_activation_sql,
)
from exasol.nb_connector.model_installation import ensure_model_subdir_config_value
from exasol.nb_connector.secret_store import Secrets

LANGUAGE_ALIAS = "PYTHON3_TE"

LATEST_KNOWN_VERSION = version("exasol_transformers_extension")

# Activation SQL for the Transformers Extension will be saved in the secret
# store with this key.
ACTIVATION_KEY = ACTIVATION_KEY_PREFIX + "te"

# The name of the connection object with BucketFS location and credentials
# will be prefixed with this string.
BFS_CONNECTION_PREFIX = "TE_BFS"

# The name of the connection object with a Huggingface token will be prefixed
# with this string.
HF_CONNECTION_PREFIX = "TE_HF"

# Models downloaded from the Huggingface archive to a local drive will be
# cached in this directory.
MODELS_CACHE_DIR = "models_cache"


def deploy_scripts(conf: Secrets, language_alias: str) -> None:
    """
    Deploys all the extension's scripts to the database.

    Parameters:
        conf:
            The secret store. The store should contain the language activation
            SQL command. This command should be created during the language
            container deployment. The store should also have the DB schema.
        language_alias:
            The language alias of the extension's language container.
    """

    with open_pyexasol_connection(conf, compression=True) as conn:
        # First need to activate the language container at the session level, otherwise the script creation fails.
        activation_sql = get_activation_sql(conf)
        conn.execute(activation_sql)

        scripts_deployer = ScriptsDeployer(
            language_alias, conf.get(CKey.db_schema), conn, install_all_scripts=True
        )
        scripts_deployer.deploy_scripts()


def initialize_te_extension(
    conf: Secrets,
    version: str = LATEST_KNOWN_VERSION,
    language_alias: str = LANGUAGE_ALIAS,
    run_deploy_container: bool = True,
    run_deploy_scripts: bool = True,
    run_encapsulate_bfs_credentials: bool = True,
    run_encapsulate_hf_token: bool = True,
    allow_override: bool = True,
) -> None:
    """
    Performs all necessary operations to get the Transformers Extension
    up and running. See the "Getting Started" and "Setup" sections of the
    extension's User Guide for details.

    Parameters:
        conf:
            The secret store. The store should contain all the required
            parameters for accessing the database and BucketFS.
        version:
            Transformers Extension version. If not specified the hardcoded
            latest known version will be used.
        language_alias:
            The language alias of the extension's language container. Normally
            this parameter would only be used for testing.
        run_deploy_container:
            If set to False will skip the language container deployment.
        run_deploy_scripts:
            If set to False will skip the deployment of the UDF scripts.
        run_encapsulate_bfs_credentials:
            If set to False will skip the creation of the database connection
            object encapsulating the BucketFS credentials.
        run_encapsulate_hf_token:
            If set to False will skip the creation of the database connection
            object encapsulating the Huggingface token.
        allow_override:
            If True allows overriding the language definition. Otherwise, if
            the database already has a language definition for the specified
            language alias, an attempt to deploy the container will result
            in a RuntimeError.
    """

    # Make the connection object names
    db_user = str(conf.get(CKey.db_user))
    token = conf.get(CKey.huggingface_token)
    hf_conn_name = "_".join([HF_CONNECTION_PREFIX, db_user]) if token else ""

    if run_deploy_container:
        container_url = TeLanguageContainerDeployer.SLC_URL_FORMATTER.format(
            version=version
        )
        deploy_language_container(
            conf,
            container_url=container_url,
            container_name=TeLanguageContainerDeployer.SLC_NAME,
            language_alias=language_alias,
            activation_key=ACTIVATION_KEY,
            path_in_bucket=PATH_IN_BUCKET_FOR_SLC,
            allow_override=allow_override,
        )

    ensure_bfs_connection(conf)
    ensure_model_subdir_config_value(conf)

    # Create the required objects in the database
    if run_deploy_scripts:
        deploy_scripts(conf, language_alias)
    if token and run_encapsulate_hf_token:
        encapsulate_huggingface_token(conf, hf_conn_name)

    # Save the connection object name in the secret store.
    conf.save(CKey.te_hf_connection, hf_conn_name)
    # Save the directory names in the secret store
    conf.save(CKey.te_models_cache_dir, MODELS_CACHE_DIR)


def upload_model_from_cache(conf: Secrets, model_name: str, cache_dir: str) -> None:
    """
    Uploads model previously downloaded and cached on a local drive. This,
    for instance, could have been done with the following code.

    from transformers import AutoTokenizer, AutoModel
    AutoTokenizer.from_pretrained(model_name, cache_dir=cache_dir)
    AutoModel.from_pretrained(model_name, cache_dir=cache_dir)

    Parameters:
        conf:
            The secret store.
        model_name:
            Name of the model at the Huggingface archive.
        cache_dir:
            Directory on the local drive where the model was cached. Each model
            should have its own cache directory.
    """

    raise NotImplementedError(
        "Uploading the model is temporarily unavailable. "
        "Awaiting changes in the Transformers Extension module."
    )


def upload_model(conf: Secrets, model_name: str, cache_dir: str, **kwargs) -> None:
    """
    Uploads model from the Huggingface hub or from the local cache in case it
    has already been downloaded from the hub. The user token, if found in the
    secret store will be passed to the Huggingface interface.

    Parameters:
        conf:
            The secret store.
        model_name:
            Name of the model at the Huggingface archive.
        cache_dir:
            Directory on the local drive where the model is to be cached.
            Each model should have its own cache directory.
        kwargs:
            Additional parameters to be passed to the `from_pretrained`
            methods of the AutoTokenizer and AutoModel. The user token, if specified
            here, will be used instead of the one in the secret store.
    """
    from transformers import (
        AutoModel,
        AutoTokenizer,
    )

    if "token" not in kwargs:
        token = conf.get(CKey.huggingface_token)
        if token:
            kwargs["token"] = token

    AutoTokenizer.from_pretrained(model_name, cache_dir=cache_dir, **kwargs)
    AutoModel.from_pretrained(model_name, cache_dir=cache_dir, **kwargs)

    upload_model_from_cache(conf, model_name, cache_dir)



import click


@click.group()
def cli():
    """
    Manage application configuration data in the Secure Configuration
    Storage (SCS).

    You can set environment variables SCS_FILE and SCS_MASTER_PASSWORD
    specifying the file containing the SCS and the master password for
    accessing and encrypting the file, respectively.

    Otherwise the scs commands will require the SCS file to be passed as
    positional argument and the master password to be typed interactively.
    """
    pass



import click

from exasol.nb_connector.ai_lab_config import Accelerator
from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import ScsOption

EXTRA_DOCKER_DB_OPTIONS = [
    ScsOption(
        "--db-mem-size",
        type=int,
        metavar="GiB",
        default=8,
        help="Database memory size (GiB)",
        scs_key=CKey.mem_size,
    ),
    ScsOption(
        "--db-disk-size",
        metavar="GiB",
        type=int,
        default=2,
        help="Database disk size (GiB)",
        scs_key=CKey.disk_size,
    ),
    ScsOption(
        "--accelerator",
        type=click.Choice(Accelerator, case_sensitive=False),
        default="none",
        help="Hardware acceleration",
        scs_key=CKey.accelerator,
    ),
]



from exasol.nb_connector.cli.options.bucketfs import BUCKETFS_OPTIONS
from exasol.nb_connector.cli.options.common import (
    COMMON_OPTIONS,
    SAVE_OPTIONS,
    SCS_OPTIONS,
)
from exasol.nb_connector.cli.options.docker_db import EXTRA_DOCKER_DB_OPTIONS
from exasol.nb_connector.cli.options.onprem import ONPREM_DB_OPTIONS
from exasol.nb_connector.cli.options.saas import EXTRA_SAAS_OPTIONS
from exasol.nb_connector.cli.options.ssl import SSL_OPTIONS


def _wrap(options):
    return SCS_OPTIONS + SAVE_OPTIONS + options + COMMON_OPTIONS


DOCKER_DB_OPTIONS = _wrap(EXTRA_DOCKER_DB_OPTIONS)

SAAS_OPTIONS = _wrap(EXTRA_SAAS_OPTIONS + SSL_OPTIONS)

ONPREM_OPTIONS = _wrap(ONPREM_DB_OPTIONS + BUCKETFS_OPTIONS + SSL_OPTIONS)



import logging
from pathlib import Path

from git import Repo

from exasol.nb_connector.slc import constants

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class GitAccess:
    @staticmethod
    def clone_from_recursively(url: str, path: Path, branch: str) -> None:
        LOG.info(f"Cloning into {path}...")
        repo = Repo.clone_from(
            constants.SLC_GITHUB_REPO,
            path,
            branch=branch,
        )
        LOG.info("Fetching submodules...")
        repo.submodule_update(recursive=True)

    @staticmethod
    def checkout_recursively(path: Path) -> None:
        repo = Repo(path)
        repo.git.checkout("--recurse-submodules", ".")

    @staticmethod
    def checkout_file(repo_path: Path, file: Path) -> None:
        repo = Repo(repo_path)
        repo.git.checkout("HEAD", "--", str(file))



import ipywidgets as widgets

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.connections import get_backend
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.config.onprem import get_onprem
from exasol.nb_connector.ui.config.saas import get_saas
from exasol.nb_connector.ui.docker.docker_db import docker_db_configuration


def configure_db(conf: Secrets) -> widgets.Widget:
    """
    Creates a db configuration UI, depending on the choice of the database.
    """
    storage_backend = get_backend(conf)
    if storage_backend == StorageBackend.saas:
        return get_saas(conf)
    elif conf.get(CKey.use_itde, "True") == "True":
        return docker_db_configuration(conf)
    else:
        return get_onprem(conf)



import ipywidgets as widgets
from IPython.display import (
    Javascript,
    clear_output,
    display,
)

# Prepare to display a popup message
notify_output = widgets.Output()
display(notify_output)


@notify_output.capture()
def display_popup(message):
    clear_output()
    message = message.replace("'", '"')
    display(Javascript(f"alert('{message}')"))



from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.connections import get_backend
from exasol.nb_connector.secret_store import Secrets


class BackendSelector:
    """
    Based on an instance of Secrets (SCS) this class provides the
    following features:

    * Tell whether a particular backend is properly selected.

    * Access the properties of the selection using proper types StorageBackend
      and bool.

    * Get the user-friendly display name of the selected backend, e.g. "Docker".

    * Checks if a proposed backend conflicts with the previously made selection.
    """

    def __init__(self, scs: Secrets):
        self._scs = scs

    @property
    def backend(self) -> StorageBackend:
        return get_backend(self._scs)

    @property
    def use_itde(self) -> bool:
        return self._scs.get(CKey.use_itde, "False") == "True"

    @property
    def backend_name(self) -> str:
        if self.backend == StorageBackend.saas:
            return "SaaS"
        if self.use_itde:
            return "Docker"
        return "on-premise"

    @property
    def knows_backend(self) -> bool:
        """
        Tells whether the current backend selection unambiguously points
        to saas, onprem, or docker.
        """
        backend = self._scs.get(CKey.storage_backend)
        if not backend:
            return False
        if self.backend == StorageBackend.saas:
            return True
        use_itde = self._scs.get(CKey.use_itde)
        return use_itde is not None

    def matches(self, backend: StorageBackend, use_itde: bool) -> bool:
        """
        Tells whether the currently selected backend matches the specified
        one.  If no backend has been selected yet (property knows_backend is
        False), then the selection is rated to be matching, as well.
        """
        if not self.knows_backend:
            return True
        return backend == self.backend and use_itde == self.use_itde



from __future__ import annotations

import ssl
import warnings
from pathlib import Path
from typing import (
    Any,
)

import exasol.bucketfs as bfs
import exasol.saas.client.api_access as saas_api
import ibis
import pyexasol
import sqlalchemy
from sqlalchemy.engine.url import URL

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.utils import optional_str_to_bool


def _optional_encryption(conf: Secrets, key: CKey = CKey.db_encryption) -> bool | None:
    return optional_str_to_bool(conf.get(key))


def _extract_ssl_options(conf: Secrets) -> dict:
    """
    Extracts SSL parameters from the provided configuration.
    Returns a dictionary in the winsocket-client format
    (see https://websocket-client.readthedocs.io/en/latest/faq.html#what-else-can-i-do-with-sslopts)
    """
    sslopt: dict[str, object] = {}

    # Is server certificate validation required?
    certificate_validation = optional_str_to_bool(conf.get(CKey.cert_vld))
    if certificate_validation is not None:
        sslopt["cert_reqs"] = (
            ssl.CERT_REQUIRED if certificate_validation else ssl.CERT_NONE
        )

    # Is a bundle with trusted CAs provided?
    trusted_ca = conf.get(CKey.trusted_ca)
    if trusted_ca:
        trusted_ca_path = Path(trusted_ca)
        if trusted_ca_path.is_dir():
            sslopt["ca_cert_path"] = trusted_ca
        elif trusted_ca_path.is_file():
            sslopt["ca_certs"] = trusted_ca
        else:
            raise ValueError(f"Trusted CA location {trusted_ca} doesn't exist.")

    # Is client's own certificate provided?
    client_certificate = conf.get(CKey.client_cert)
    if client_certificate:
        if not Path(client_certificate).is_file():
            raise ValueError(f"Certificate file {client_certificate} doesn't exist.")
        sslopt["certfile"] = client_certificate
        private_key = conf.get(CKey.client_key)
        if private_key:
            if not Path(private_key).is_file():
                raise ValueError(f"Private key file {private_key} doesn't exist.")
            sslopt["keyfile"] = private_key

    return sslopt


def get_backend(conf: Secrets) -> StorageBackend:
    """
    Tries to find which backend was selected in the configuration. If the relevant
    configuration element is not there - which may be the case if the configuration
    has been created before the SaaS support was introduced - returns the
    StorageBackend.onprem.
    """
    storage_backend = conf.get(CKey.storage_backend, StorageBackend.onprem.name)
    return StorageBackend[storage_backend]  # type: ignore


def get_external_host(conf: Secrets) -> str:
    """Constructs the host part of a DB URL using provided configuration parameters."""
    return f"{conf.get(CKey.db_host_name)}:{conf.get(CKey.db_port)}"


def get_udf_bucket_path(conf: Secrets) -> str:
    """
    Builds the path of the BucketFS bucket specified in the configuration,
    as it's seen in the udf's file system.
    """
    bucket = open_bucketfs_bucket(conf)
    return bucket.udf_path


def get_saas_database_id(conf: Secrets) -> str:
    """
    Gets the SaaS database id using the available configuration elements.
    """
    saas_database_id = conf.get(CKey.saas_database_id)

    if saas_database_id:
        return saas_database_id

    return saas_api.get_database_id(
        host=conf[CKey.saas_url],
        account_id=conf[CKey.saas_account_id],
        pat=conf[CKey.saas_token],
        database_name=conf[CKey.saas_database_name],
    )


def _get_pyexasol_connection_params(conf: Secrets, **kwargs) -> dict[str, Any]:
    if get_backend(conf) == StorageBackend.onprem:
        conn_params: dict[str, Any] = {
            "dsn": get_external_host(conf),
            "user": conf.get(CKey.db_user),
            "password": conf.get(CKey.db_password),
        }
    else:
        conn_params = saas_api.get_connection_params(
            host=conf[CKey.saas_url],
            account_id=conf[CKey.saas_account_id],
            pat=conf[CKey.saas_token],
            database_id=conf.get(CKey.saas_database_id),
            database_name=conf.get(CKey.saas_database_name),
        )

    encryption = _optional_encryption(conf)
    if encryption is not None:
        conn_params["encryption"] = encryption
    ssopt = _extract_ssl_options(conf)
    if ssopt:
        conn_params["websocket_sslopt"] = ssopt

    conn_params.update(kwargs)
    return conn_params


def open_pyexasol_connection(conf: Secrets, **kwargs) -> pyexasol.ExaConnection:
    """
    Opens a pyexasol connection using provided configuration parameters.
    Supports both On-Prem and Saas backends.
    Does NOT set the default schema, even if it is defined in the configuration.

    Any additional parameters can be passed to pyexasol via the kwargs.
    Parameters in kwargs override the correspondent values in the configuration.

    The configuration should provide the following parameters:

    On-Prem:
        - Server address and port (db_host_name, db_port),
        - Client security credentials (db_user, db_password).

    Saas:
        - SaaS service url (saas_url),
        - SaaS account id (saas_account_id),
        - Database id or name (saas_database_id or saas_database_name),
        - Client security credentials (saas_token).

    Optional parameters include:
        - Secured comm flag (db_encryption),
        - Some of the SSL options (cert_vld, trusted_ca, client_cert).

    If the schema is not provided then it should be set explicitly in every SQL statement.
    For other optional parameters the default settings are as per the pyexasol interface.
    """

    conn_params = _get_pyexasol_connection_params(conf, **kwargs)
    return pyexasol.connect(**conn_params)


def open_sqlalchemy_connection(conf: Secrets):
    """
    Creates an Exasol SQLAlchemy websocket engine using provided configuration parameters.
    Supports both On-Prem and Saas backends.
    Sets the default schema if it is defined in the configuration.

    The configuration should provide the following parameters:

    On-Prem:
        - Server address and port (db_host_name, db_port),
        - Client security credentials (db_user, db_password).

    Saas:
        - SaaS service url (saas_url),
        - SaaS account id (saas_account_id),
        - Database id or name (saas_database_id or saas_database_name),
        - Client security credentials (saas_token).

    Optional parameters include:
        - Secured comm flag (db_encryption).
        - Validation of the server's TLS/SSL certificate by the client (cert_vld).

    If the schema is not provided then it should be set explicitly in every SQL statement.
    For other optional parameters the default settings are as per the Exasol SQLAlchemy interface.
    Currently, it's not possible to use a bundle of trusted CAs other than the default. Neither
    it is possible to set the client TLS/SSL certificate.
    """

    query_params = {}
    encryption = _optional_encryption(conf)
    if encryption is not None:
        query_params["ENCRYPTION"] = "Yes" if encryption else "No"
    certificate_validation = _extract_ssl_options(conf).get("cert_reqs")
    if (certificate_validation is not None) and (not certificate_validation):
        query_params["SSLCertificate"] = "SSL_VERIFY_NONE"

    if get_backend(conf) == StorageBackend.onprem:
        conn_params: dict[str, Any] = {
            "host": conf.get(CKey.db_host_name),
            "port": int(conf.get(CKey.db_port)),  # type: ignore
            "username": conf.get(CKey.db_user),
            "password": conf.get(CKey.db_password),
        }
    else:
        conn_params = saas_api.get_connection_params(
            host=conf[CKey.saas_url],
            account_id=conf[CKey.saas_account_id],
            pat=conf[CKey.saas_token],
            database_id=conf.get(CKey.saas_database_id),
            database_name=conf.get(CKey.saas_database_name),
        )
        host, port = str(conn_params["dsn"]).split(":")
        conn_params = {
            "host": host,
            "port": int(port),
            "username": conn_params["user"],
            "password": conn_params["password"],
        }

    websocket_url = URL.create(
        "exa+websocket",
        **conn_params,
        database=conf.get(CKey.db_schema),
        query=query_params,
    )

    return sqlalchemy.create_engine(websocket_url)


def _get_onprem_bucketfs_url(conf: Secrets) -> str:
    bucketfs_url_prefix = (
        "https" if _optional_encryption(conf, CKey.bfs_encryption) else "http"
    )
    bucketfs_host = conf.get(CKey.bfs_host_name, conf.get(CKey.db_host_name))
    return f"{bucketfs_url_prefix}://{bucketfs_host}:{conf.get(CKey.bfs_port)}"


def _get_ca_cert_verification(conf: Secrets) -> Any:
    sslopt = _extract_ssl_options(conf)
    verify = sslopt.get("cert_reqs") == ssl.CERT_REQUIRED
    return sslopt.get("ca_certs") or sslopt.get("ca_cert_path") or verify


def open_bucketfs_connection(conf: Secrets) -> bfs.BucketLike:
    """
    This function is deprecated, please use open_bucketfs_bucket(conf) instead.
    """

    warnings.warn(
        "open_bucketfs_connection is deprecated. Use open_bucketfs_bucket instead.",
        DeprecationWarning,
    )
    return open_bucketfs_bucket(conf)


def open_bucketfs_bucket(conf: Secrets) -> bfs.BucketLike:
    """
    Connects to a BucketFS service using provided configuration parameters.
    Returns the BucketLike object for the bucket selected in the configuration.
    Supports both On-Prem and Saas backends.

    The configuration should provide the following parameters;

    On-Prem:
        - Host name and port of the BucketFS service (bfs_host_name or db_host_name, bfs_port),
        - Client security credentials (bfs_user, bfs_password).
        - Bucket name (bfs_bucket)

    Saas:
        - SaaS service url (saas_url),
        - SaaS account id (saas_account_id),
        - Database id or name (saas_database_id or saas_database_name),
        - Client security credentials (saas_token).

    Optional parameters include:
        - Secured comm flag (bfs_encryption), defaults to False.
        - Some of the SSL options (cert_vld, trusted_ca).
    """

    if get_backend(conf) == StorageBackend.onprem:
        bucketfs_url = _get_onprem_bucketfs_url(conf)
        verify = _get_ca_cert_verification(conf)
        bucketfs_credentials = {
            conf.get(CKey.bfs_bucket): {
                "username": conf.get(CKey.bfs_user),
                "password": conf.get(CKey.bfs_password),
            }
        }

        # Connect to the BucketFS service and navigate to the bucket of choice.
        bucketfs = bfs.Service(
            bucketfs_url,
            bucketfs_credentials,  # type: ignore
            verify,  # type: ignore
            conf.get(CKey.bfs_service),
        )
        return bucketfs[conf.get(CKey.bfs_bucket)]  # type: ignore

    else:
        saas_url, saas_token, saas_account_id = (
            conf.get(key)
            for key in [CKey.saas_url, CKey.saas_token, CKey.saas_account_id]
        )
        saas_database_id = get_saas_database_id(conf)
        return bfs.SaaSBucket(
            url=saas_url,  # type: ignore
            account_id=saas_account_id,  # type: ignore
            database_id=saas_database_id,  # type: ignore
            pat=saas_token,  # type: ignore
        )


def open_bucketfs_location(conf: Secrets) -> bfs.path.PathLike:
    """
    Similar to `open_buckets_connection`, but returns a PathLike interface.
    """
    if get_backend(conf) == StorageBackend.onprem:
        return bfs.path.build_path(
            backend=bfs.path.StorageBackend.onprem,
            url=_get_onprem_bucketfs_url(conf),
            username=conf.get(CKey.bfs_user),
            password=conf.get(CKey.bfs_password),
            verify=_get_ca_cert_verification(conf),
            bucket_name=conf.get(CKey.bfs_bucket),
            service_name=conf.get(CKey.bfs_service),
        )
    else:
        return bfs.path.build_path(
            backend=bfs.path.StorageBackend.saas,
            url=conf.get(CKey.saas_url),
            account_id=conf.get(CKey.saas_account_id),
            database_id=get_saas_database_id(conf),
            pat=conf.get(CKey.saas_token),
        )


def open_ibis_connection(conf: Secrets, **kwargs):
    """
    Creates a connection to Ibis with Exasol backend.

    The parameters are similar to those of open_pyexasol_connection function.
    The downstream call signature is also similar to pyexasol, except that the dsn is
    provided in two separate parts - host and port.

    Unlike open_pyexasol_connection, this function sets the default schema if it's
    defined in the configuration.
    """

    conn_params = _get_pyexasol_connection_params(conf, **kwargs)

    dsn = conn_params.pop("dsn")
    host_port = dsn.split(":")
    conn_params["host"] = host_port[0]
    if len(host_port) > 1:
        conn_params["port"] = int(host_port[1])

    schema = conf.get(CKey.db_schema)
    if schema:
        conn_params["schema"] = schema

    return ibis.exasol.connect(**conn_params)



import os
from dataclasses import dataclass
from typing import Any

from exasol.ai.text.extractors.bucketfs_model_repository import BucketFSRepository
from exasol_transformers_extension.utils.bucketfs_model_specification import (
    BucketFSModelSpecification,
)
from exasol_transformers_extension.utils.model_utils import install_huggingface_model
from yaspin import yaspin

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.bfs_connection import (
    PATH_IN_BUCKET,
    ensure_bfs_connection,
    ensure_bfs_connection_name,
)
from exasol.nb_connector.connections import open_bucketfs_location
from exasol.nb_connector.secret_store import Secrets

CHECKMARK = "\u2705"
"""
Checkmark symbol for signalling success after an operation using an
animated spinner from https://github.com/pavdmyt/yaspin.
"""

# Models will be uploaded into this directory in BucketFS.
DEF_BFS_MODELS_DIR = "models"


@dataclass
class TransformerModel:
    name: str
    task_type: str
    factory: Any


def _interactive_usage() -> bool:
    return os.environ.get("INTERACTIVE", "True").lower() == "true"


def ensure_model_subdir_config_value(conf: Secrets) -> str:
    if value := conf.get(CKey.bfs_model_subdir):
        return value
    conf.save(CKey.bfs_model_subdir, DEF_BFS_MODELS_DIR)
    return DEF_BFS_MODELS_DIR


def install_model(conf: Secrets, model: TransformerModel) -> None:
    """
    Download and install the specified Huggingface model.
    """
    ensure_bfs_connection(conf)
    bucketfs_location = open_bucketfs_location(conf) / PATH_IN_BUCKET
    with yaspin(text=f"- Huggingface model {model.name}") as spinner:
        if not _interactive_usage():
            spinner.hide()
        sub_dir = ensure_model_subdir_config_value(conf)
        model_spec = BucketFSModelSpecification(
            model.name, model.task_type, "", sub_dir
        )
        install_huggingface_model(
            bucketfs_location=bucketfs_location,
            model_spec=model_spec,
            model_factory=model.factory,
        )
    spinner.ok(CHECKMARK)


def create_model_repository(conf: Secrets) -> BucketFSRepository:
    """
    Creates a BucketFSRepository encapsulating using the sub-directory from the secret store.
    Parameters:
         conf:
            The secret store.
    """
    ensure_bfs_connection(conf)
    return BucketFSRepository(
        connection_name=ensure_bfs_connection_name(conf),
        sub_dir=ensure_model_subdir_config_value(conf),
    )



class SlcError(Exception):
    """
    Signals errors related to ScriptLanguageContainer:

    * The name of the SLC is not unique

    * the Secure Configuration Storage (SCS / secrets / conf) does not contain
      a required option

    * The SLC Git repository has not been checked out (cloned)
    """



import ipywidgets as widgets

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.config.generic import get_config

DEFAULT_SCHEMA = "Default Schema"


def get_onprem(conf: Secrets) -> widgets.Widget:
    """
    Creates a UI form for editing an external Exasol Database configuration.
    """

    inputs = [
        [
            (
                "Host Name",
                widgets.Text(value=conf.get(CKey.db_host_name, "localhost")),
                CKey.db_host_name,
            ),
            (
                "Port",
                widgets.IntText(value=int(conf.get(CKey.db_port) or 8563)),
                CKey.db_port,
            ),
            ("User Name", widgets.Text(value=conf.get(CKey.db_user)), CKey.db_user),
            (
                "Password",
                widgets.Password(value=conf.get(CKey.db_password)),
                CKey.db_password,
            ),
            (
                DEFAULT_SCHEMA,
                widgets.Text(value=conf.get(CKey.db_schema, "AI_LAB")),
                CKey.db_schema,
            ),
            (
                "Encrypted Comm.",
                widgets.Checkbox(
                    value=conf.get(CKey.db_encryption, "True") == "True", indent=False
                ),
                CKey.db_encryption,
            ),
        ],
        [
            (
                "External Host Name",
                widgets.Text(value=conf.get(CKey.bfs_host_name, "localhost")),
                CKey.bfs_host_name,
            ),
            (
                "Internal Host Name",
                widgets.Text(value=conf.get(CKey.bfs_host_name, "localhost")),
                CKey.bfs_internal_host_name,
            ),
            (
                "External Port",
                widgets.IntText(value=int(conf.get(CKey.bfs_port) or 2580)),
                CKey.bfs_port,
            ),
            (
                "Internal Port",
                widgets.IntText(value=int(conf.get(CKey.bfs_port) or 2580)),
                CKey.bfs_internal_port,
            ),
            ("User Name", widgets.Text(value=conf.get(CKey.bfs_user)), CKey.bfs_user),
            (
                "Password",
                widgets.Password(value=conf.get(CKey.bfs_password)),
                CKey.bfs_password,
            ),
            (
                "Service Name",
                widgets.Text(value=conf.get(CKey.bfs_service, "bfsdefault")),
                CKey.bfs_service,
            ),
            (
                "Bucket Name",
                widgets.Text(value=conf.get(CKey.bfs_bucket, "default")),
                CKey.bfs_bucket,
            ),
            (
                "Encrypted Comm.",
                widgets.Checkbox(
                    value=conf.get(CKey.bfs_encryption, "True") == "True", indent=False
                ),
                CKey.bfs_encryption,
            ),
        ],
        [
            (
                "Validate Certificate",
                widgets.Checkbox(
                    value=conf.get(CKey.cert_vld, "True") == "True", indent=False
                ),
                CKey.cert_vld,
            ),
            (
                "Trusted CA File/Dir",
                widgets.Text(value=conf.get(CKey.trusted_ca)),
                CKey.trusted_ca,
            ),
            (
                "Certificate File",
                widgets.Text(value=conf.get(CKey.client_cert)),
                CKey.client_cert,
            ),
            (
                "Private Key File",
                widgets.Text(value=conf.get(CKey.client_key)),
                CKey.client_key,
            ),
        ],
    ]

    group_names = [
        "Database Connection",
        "BucketFS Connection",
        "TLS/SSL Configuration",
    ]

    return get_config(conf, inputs, group_names)



import os
from enum import Enum

import ipywidgets as widgets
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    TaskFailures,
    TaskRuntimeError,
)

from exasol.nb_connector.ai_lab_config import Accelerator
from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.itde_manager import (
    ItdeContainerStatus,
    bring_itde_up,
    get_itde_status,
    restart_itde,
    take_itde_down,
)
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.common.popup_message import display_popup
from exasol.nb_connector.ui.common.ui_styles import config_styles
from exasol.nb_connector.ui.common.useful_urls import Urls
from exasol.nb_connector.ui.config.generic import get_config

DEFAULT_SCHEMA = "Default Schema"


class DockerDbDisplayStatus(Enum):
    """User-facing status messages for the Exasol Docker-DB."""

    READY = "Exasol Docker-DB is READY"
    STOPPED = "Exasol Docker-DB is STOPPED"
    DISCONNECTED = "Exasol Docker-DB is NOT CONNECTED"
    MISSING = "Exasol Docker-DB is NOT CREATED"
    INACCESSIBLE = "Exasol Docker-DB is INACCESSIBLE"


def docker_db_configuration(conf: Secrets) -> widgets.Widget:
    """
    Creates a UI form for editing the Exasol Docker-DB configuration.
    """

    inputs = [
        [
            (
                "Memory Size (GiB)",
                widgets.IntText(value=int(conf.get(CKey.mem_size) or 2)),
                CKey.mem_size,
            ),
            (
                "Disk Size (GiB)",
                widgets.IntText(value=int(conf.get(CKey.disk_size) or 2)),
                CKey.disk_size,
            ),
            (
                "Accelerator",
                widgets.Dropdown(
                    value=conf.get(CKey.accelerator, "none"),
                    options=[x.value for x in Accelerator],
                ),
                CKey.accelerator,
            ),
            (
                DEFAULT_SCHEMA,
                widgets.Text(value=conf.get(CKey.db_schema, "AI_LAB")),
                CKey.db_schema,
            ),
        ]
    ]

    group_names = ["Database Configuration"]

    return get_config(conf, inputs, group_names)


def _itde_error_message(header: str, e: Exception) -> str:
    if isinstance(e, TaskRuntimeError):
        if isinstance(e.__cause__, TaskFailures):
            err_message = f"{e.msg} {e.__cause__}"
        elif e.inner:
            causes = "\n".join(e.inner)
            err_message = f"{e.msg} This was caused by\n{causes}"
        else:
            err_message = e.msg
    else:
        err_message = str(e)
    return f"{header}: {err_message}"


def _get_docker_db_action_buttons(
    conf: Secrets, itde_exists: bool, itde_ready: bool, display_status: widgets.Widget
) -> list[widgets.Button]:
    """
    Creates one or two action buttons with the correspondent on_click functions for managing the
    Exasol Docker-DB. Depending on the current status (idte_exists, itde_ready) of the docker
    container, the "Start", "Restart" or both buttons are created.
    When the action is completed successfully, the running status is displayed in the provided
    widget (display_status).
    """

    def start_docker_db(btn):
        try:
            # Need to check if the Exasol Docker-DB still exists and not running because
            # the situation might have changed while the widgets were hanging around.
            itde_status_now = get_itde_status(conf)
            if itde_status_now != ItdeContainerStatus.READY:
                if itde_status_now == ItdeContainerStatus.ABSENT:
                    bring_itde_up(conf)
                else:
                    restart_itde(conf)
            # Indicate the successful completion.
            display_status.value = DockerDbDisplayStatus.READY.value
            btn.icon = "check"
        except Exception as e:
            display_popup(
                _itde_error_message("Failed to start the Exasol Docker-DB", e)
            )

    def restart_docker_db(btn):
        try:
            # Check if the Docker-DB state has changed since initial display of the UI widgets.
            itde_status_now = get_itde_status(conf)
            if itde_status_now != ItdeContainerStatus.ABSENT:
                take_itde_down(conf)
            bring_itde_up(conf)
            # Indicate the successful completion.
            display_status.value = DockerDbDisplayStatus.READY.value
            btn.icon = "check"
        except Exception as e:
            display_popup(
                _itde_error_message("Failed to restart the Exasol Docker-DB:", e)
            )

    if itde_ready:
        btn_restart = widgets.Button(description="Recreate and Start")
        btn_restart.on_click(restart_docker_db)
        return [btn_restart]
    elif itde_exists:
        btn_start = widgets.Button(description="Start")
        btn_start.on_click(start_docker_db)
        btn_restart = widgets.Button(description="Recreate and Start")
        btn_restart.on_click(restart_docker_db)
        return [btn_start, btn_restart]
    else:
        btn_start = widgets.Button(description="Create and Start")
        btn_start.on_click(start_docker_db)
        return [btn_start]


def _create_warning(warning_text: str) -> widgets.Widget:
    return widgets.HTML(
        value="<style>p.itde_warning{word-wrap: break-word; color: red;}</style>"
        '<p class="itde_warning">' + warning_text + " </p>"
    )


def manage_docker_db(conf: Secrets, socket="/var/run/docker.sock") -> widgets.Widget:
    """
    A UI for starting or restarting the Exasol Docker-DB.
    It checks if an instance of the Exasol Docker-DB is already running or if it exists.
    In that case a warning is displayed.
    """

    if conf.get(CKey.use_itde) != "True":
        return None

    ui_look = config_styles()

    # Check if the docker-socket has been mounted
    socket_mounted = os.path.exists(socket)

    header_lbl = widgets.Label(style=ui_look.header_style, layout=ui_look.header_layout)
    group_items = [header_lbl]

    if socket_mounted:
        # Get and display the current status of the Exasol Docker-DB.
        itde_status = get_itde_status(conf)
        if itde_status == ItdeContainerStatus.READY:
            header_lbl.value = DockerDbDisplayStatus.READY.value
        elif itde_status == ItdeContainerStatus.RUNNING:
            header_lbl.value = DockerDbDisplayStatus.DISCONNECTED.value
        elif itde_status == ItdeContainerStatus.STOPPED:
            header_lbl.value = DockerDbDisplayStatus.STOPPED.value
        else:
            header_lbl.value = DockerDbDisplayStatus.MISSING.value

        # Add a warning message about recreating an existing Exasol Docker-DB.
        itde_exists = itde_status != ItdeContainerStatus.ABSENT
        itde_ready = itde_status == ItdeContainerStatus.READY
        if itde_exists:
            warning_text = (
                "Please note that recreating the Exasol Docker-DB will result in the loss of all data stored in the "
                f'{"running" if itde_ready else "existing"} instance of the database!'
            )
            warning_html = _create_warning(warning_text)
            group_items.append(widgets.Box([warning_html], layout=ui_look.row_layout))

        # Create action buttons.
        action_buttons = _get_docker_db_action_buttons(
            conf, itde_exists, itde_ready, header_lbl
        )
        for btn in action_buttons:
            btn.style = ui_look.button_style
            btn.layout = ui_look.button_layout
    else:

        header_lbl.value = DockerDbDisplayStatus.INACCESSIBLE.value
        warning_text = (
            f"The docker socket is not mounted. Please consult the "
            f'<a href="{Urls.user_manual_docker_db.value}" target="_blank">documentation</a> '
            "on how to start the AI-Lab that will use the Integrated Exasol Docker-DB."
        )
        warning_html = _create_warning(warning_text)
        group_items.append(widgets.Box([warning_html], layout=ui_look.row_layout))
        action_buttons = []

    # Put all UI elements together.
    items = [
        widgets.Box(group_items, layout=ui_look.group_layout),
        widgets.Box(action_buttons, layout=ui_look.row_layout),
    ]
    ui = widgets.Box(items, layout=ui_look.outer_layout)
    return ui



from __future__ import annotations

import logging
import re
from pathlib import (
    Path,
)
from typing import (
    Any,
)

import requests
from exasol.slc import api as exaslct_api
from exasol.slc.api.get_language_definition_builder import (
    get_language_definition_builder,
)
from exasol.slc.models.compression_strategy import CompressionStrategy
from exasol_integration_test_docker_environment.lib.docker import (
    ContextDockerClient,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.slc import constants
from exasol.nb_connector.slc.git_access import GitAccess
from exasol.nb_connector.slc.package_file_editor import append_packages
from exasol.nb_connector.slc.package_types import (
    CondaPackageDefinition,
    PipPackageDefinition,
)
from exasol.nb_connector.slc.slc_compression_strategy import SlcCompressionStrategy
from exasol.nb_connector.slc.slc_flavor import (
    SlcError,
    SlcFlavor,
)
from exasol.nb_connector.slc.workspace import (
    Workspace,
    current_directory,
)
from exasol.nb_connector.utils import optional_str_to_bool

NAME_PATTERN = re.compile(r"^[A-Z][A-Z0-9_]*$", flags=re.IGNORECASE)


def _verify_name(slc_name: str) -> None:
    if not NAME_PATTERN.match(slc_name):
        raise SlcError(
            f'SLC name "{slc_name}" doesn\'t match'
            f' regular expression "{NAME_PATTERN}".'
        )


class ScriptLanguageContainer:
    """
    Support building different flavors of Exasol Script Language
    Containers (SLCs) using the SLCT.

    Parameter ``name`` serves as base of the language alias and a key for the related flavor stored in the
    Secure Configuration Storage (SCS / secrets / conf).  The flavor is used
    as a template for building the SLC.

    If the flavor is missing in the SCS or the SLC Git repository has not been
    checked out (i.e. cloned) into the checkout_dir, then the constructor will
    raise an SlcError.

    Additionally, the caller needs to ensure, that a flavor with this name is
    contained in the SLC release specified in variable
    constants.SLC_RELEASE_TAG.
    """

    def __init__(
        self,
        secrets: Secrets,
        name: str,
    ):
        self.secrets = secrets
        self.name = name
        _verify_name(name)
        self.flavor = SlcFlavor(name).verify(secrets)
        self.compression_strategy = SlcCompressionStrategy(name).verify(secrets)
        self.workspace = Workspace.for_slc(name, secrets)
        if not self.checkout_dir.is_dir():
            raise SlcError(
                f"SLC Git repository not checked out to {self.checkout_dir}."
            )
        if not self.flavor_path.is_dir():
            raise SlcError(
                f"Given flavor {self.flavor} not found in version {constants.SLC_RELEASE_TAG} of "
                "Script-Languages-Release. "
                "Check out available flavors at https://github.com/exasol/script-languages-release/releases/tag/{constants.SLC_RELEASE_TAG}."
            )

    @classmethod
    def create(
        cls,
        secrets: Secrets,
        name: str,
        flavor: str,
        compression_strategy: CompressionStrategy = CompressionStrategy.GZIP,
    ) -> ScriptLanguageContainer:
        _verify_name(name)
        slc_flavor = SlcFlavor(name)
        if slc_flavor.exists(secrets):
            raise SlcError(
                "Secure Configuration Storage already contains a"
                f" flavor for SLC name {name}."
            )
        slc_compression_strategy = SlcCompressionStrategy(slc_name=name)
        if slc_compression_strategy.exists(secrets):
            raise SlcError(
                "Secure Configuration Storage already contains a"
                f" compression strategy for SLC name {name}."
            )
        slc_flavor.save(secrets, flavor)
        slc_compression_strategy.save(secrets, compression_strategy)
        workspace = Workspace.for_slc(name, secrets)
        workspace.clone_slc_repo()
        return cls(secrets=secrets, name=name)

    @classmethod
    def create_or_open(
        cls,
        secrets: Secrets,
        name: str,
        flavor: str,
        compression_strategy: CompressionStrategy = CompressionStrategy.GZIP,
    ) -> ScriptLanguageContainer:
        _verify_name(name)
        slc_flavor = SlcFlavor(name)
        if slc_flavor.exists(secrets):
            logging.info(
                "Secure Configuration Storage already contains a"
                " flavor for SLC name %s.",
                name,
            )
        else:
            slc_flavor.save(secrets, flavor)

        slc_compression_strategy = SlcCompressionStrategy(slc_name=name)
        if slc_compression_strategy.exists(secrets):
            logging.info(
                "Secure Configuration Storage already contains a"
                f" compression strategy for SLC name {name}."
            )
        else:
            slc_compression_strategy.save(secrets, compression_strategy)
        workspace = Workspace.for_slc(name, secrets)
        workspace.clone_slc_repo()
        return cls(secrets=secrets, name=name)

    @property
    def language_alias(self) -> str:
        """
        Is case-insensitive.
        """
        return f"custom_slc_{self.name}"

    @property
    def checkout_dir(self) -> Path:
        return self.workspace.git_clone_path

    @property
    def _flavor_path_rel(self) -> str:
        return str(self.flavor_path.relative_to(self.checkout_dir))

    @property
    def flavor_path(self) -> Path:
        return self.checkout_dir / constants.FLAVORS_PATH_IN_SLC_REPO / self.flavor

    @property
    def custom_packages_dir(self):
        """
        Returns the path to the custom packages directory of the flavor
        """
        return self.flavor_path / "flavor_customization" / "packages"

    @property
    def custom_pip_file(self) -> Path:
        """
        Returns the path to the custom pip packages file of the flavor
        """
        return self.custom_packages_dir / "python3_pip_packages"

    @property
    def custom_conda_file(self) -> Path:
        """
        Returns the path to the custom conda packages file of the flavor
        """
        return self.custom_packages_dir / "conda_packages"

    def restore_custom_pip_file(self):
        """
        Restores the custom pip packages file from Git. All changes will be overwritten.
        """
        GitAccess.checkout_file(
            self.checkout_dir / "script-languages",
            self.custom_pip_file.relative_to(self.checkout_dir),
        )

    def restore_custom_conda_file(self):
        """
        Restores the custom conda packages file from Git. All changes will be overwritten.
        """
        GitAccess.checkout_file(
            self.checkout_dir / "script-languages",
            self.custom_conda_file.relative_to(self.checkout_dir),
        )

    def export(self) -> None:
        """
        Exports the current SLC to the export directory.
        """
        with current_directory(self.checkout_dir):
            exaslct_api.export(
                flavor_path=(str(self._flavor_path_rel),),
                export_path=str(self.workspace.export_path),
                output_directory=str(self.workspace.output_path),
                release_name=self.language_alias,
                compression_strategy=self.compression_strategy,
            )

    def export_no_copy(self) -> None:
        """
        Exports the current SLC to the internal output directory only, without copying to the export directory.
        """
        with current_directory(self.checkout_dir):
            exaslct_api.export(
                flavor_path=(str(self._flavor_path_rel),),
                output_directory=str(self.workspace.output_path),
                release_name=self.language_alias,
                compression_strategy=self.compression_strategy,
            )

    def _generate_bfs_params_from_secret_store(self):
        bfs_params: dict[str, Any] = {
            k: self.secrets.get(v)
            for k, v in [
                ("bucketfs_host", CKey.bfs_host_name),
                ("bucketfs_port", CKey.bfs_port),
                ("bucketfs_user", CKey.bfs_user),
                ("bucketfs_password", CKey.bfs_password),
                ("bucketfs_name", CKey.bfs_service),
                ("bucket", CKey.bfs_bucket),
            ]
        }

        bucketfs_use_https = optional_str_to_bool(self.secrets.get(CKey.bfs_encryption))
        if bucketfs_use_https is not None:
            bfs_params["bucketfs_use_https"] = bucketfs_use_https
        use_ssl_cert_validation = optional_str_to_bool(self.secrets.get(CKey.cert_vld))
        if use_ssl_cert_validation is not None:
            bfs_params["use_ssl_cert_validation"] = use_ssl_cert_validation
        ssl_cert_path = self.secrets.get(CKey.trusted_ca)
        if ssl_cert_path is not None:
            bfs_params["ssl_cert_path"] = ssl_cert_path
        return bfs_params

    def deploy(self):
        """
        Deploys the current script-languages-container to the database and
        stores the activation string in the Secure Configuration Storage.
        """
        bfs_params = self._generate_bfs_params_from_secret_store()

        with current_directory(self.checkout_dir):
            result = exaslct_api.deploy(
                flavor_path=(str(self._flavor_path_rel),),
                **bfs_params,
                path_in_bucket=constants.PATH_IN_BUCKET,
                release_name=self.language_alias,
                output_directory=str(self.workspace.output_path),
                compression_strategy=self.compression_strategy,
            )
            deploy_result = result[self._flavor_path_rel]["release"]
            builder = deploy_result.language_definition_builder
            components = builder.generate_definition_components()
            builder.add_custom_alias(components[0].alias, self.language_alias)
            lang_def = builder.generate_definition()
            self.secrets.save(self._alias_key, lang_def)

    @property
    def _alias_key(self):
        return constants.SLC_ACTIVATION_KEY_PREFIX + self.language_alias

    @property
    def activation_key(self) -> str:
        """
        Returns the language activation string for the uploaded script-language-container.
        Can be used in `ALTER SESSION` or `ALTER_SYSTEM` SQL commands to activate
        the language of the uploaded script-language-container.
        """
        try:
            return self.secrets[self._alias_key]
        except AttributeError as ex:
            raise SlcError(
                "Secure Configuration Storage does not contains an activation key."
            ) from ex

    def generate_activation_key(self, add_to_secret_store: bool) -> str:
        """
        Generates the language activation string for the uploaded script-language-container.
        Can be used in `ALTER SESSION` or `ALTER_SYSTEM` SQL commands to activate
        the language of the uploaded script-language-container, for example, if you want to use the SLC in your SQL Editor.
        Furthermore, this method can be used as fallback if the deploy-function failed for some reason,
        but the script-language-container was uploaded to the BucketFS.
        Then it will register the Language activation for this SLC in the Secure Configuration Store.
        """
        bfs_params = self._generate_bfs_params_from_secret_store()

        with current_directory(self.checkout_dir):
            builder = get_language_definition_builder(
                flavor_path=str(self._flavor_path_rel),
                bucketfs_name=bfs_params["bucketfs_name"],
                bucket_name=bfs_params["bucket"],
                path_in_bucket=constants.PATH_IN_BUCKET,
                container_name=f"{self.flavor}-release-{self.language_alias}",  # Currently this can't be retrieved from exaslct. Need to use a hard coded value here.
            )
            components = builder.generate_definition_components()
            builder.add_custom_alias(components[0].alias, self.language_alias)
            lang_def = builder.generate_definition()
            if add_to_secret_store:
                self.secrets.save(self._alias_key, lang_def)
            return lang_def

    def append_custom_pip_packages(self, pip_packages: list[PipPackageDefinition]):
        """
        Appends packages to the custom pip packages file.
        Note: This method is not idempotent: Multiple calls with the same
        package definitions will result in duplicated entries.
        """
        append_packages(self.custom_pip_file, PipPackageDefinition, pip_packages)

    def append_custom_conda_packages(
        self, conda_packages: list[CondaPackageDefinition]
    ):
        """
        Appends packages to the custom conda packages file.
        Note: This method is not idempotent: Multiple calls with the same
        package definitions will result in duplicated entries.
        """
        append_packages(self.custom_conda_file, CondaPackageDefinition, conda_packages)

    @property
    def docker_image_tags(self) -> list[str]:
        """
        Return list of Docker image tags related to the current SLC.
        """
        image_name = constants.SLC_DOCKER_IMG_NAME
        prefix = f"{image_name}:{self.flavor}"
        with ContextDockerClient() as docker_client:
            images = docker_client.images.list(name=image_name)
            return [tag for img in images if (tag := img.tags[0]).startswith(prefix)]

    @classmethod
    def clean_docker_images(cls, output_path: Path = Path.cwd()) -> None:
        """
        Deletes all SLC related local docker images.
        :param output_path: Defines the path where the log files are stored.
        """
        exaslct_api.clean_all_images(
            output_directory=str(output_path),
        )

    @classmethod
    def list_available_flavors(cls) -> list[str]:
        owner = "exasol"
        repo = "script-languages-release"
        path = "flavors"

        url = f"https://api.github.com/repos/{owner}/{repo}/contents/{path}"
        params = {"ref": constants.SLC_RELEASE_TAG}

        try:
            response = requests.get(url, params=params, timeout=30)
            data = response.json()
            return [
                (item["name"])
                for item in data
                if item["type"] in ("dir", "symlink")
                and item["name"].startswith("template")
            ]
        except requests.exceptions.RequestException as ex:
            raise SlcError("Unable to fetch flavor definitions.") from ex



import sys
from pathlib import Path

import click

from exasol.nb_connector.cli import reporting as report
from exasol.nb_connector.cli.groups import cli
from exasol.nb_connector.cli.options import SCS_OPTIONS
from exasol.nb_connector.cli.param_wrappers import add_params
from exasol.nb_connector.cli.processing import processing
from exasol.nb_connector.cli.processing.option_set import ScsCliError


@cli.command()
@add_params(SCS_OPTIONS)
@click.option(
    "--connect/--no-connect",
    is_flag=True,
    help="""Verify if connecting to the configured Exasol database instance
    succeeds.""",
)
def check(scs_file: Path, connect: bool):
    """
    Check the configuration currently saved to the Secure Configuration
    Storage and verify if it contains all required parameters.

    Optionally also verify if a connection to the configured Exasol database
    instance is successful.
    """
    try:
        processing.check_scs(scs_file, connect)
    except ScsCliError as ex:
        report.error(ex)
        sys.exit(1)



from pathlib import Path

from exasol.nb_connector.ai_lab_config import (
    StorageBackend,
)
from exasol.nb_connector.cli.groups import cli
from exasol.nb_connector.cli.options import (
    DOCKER_DB_OPTIONS,
    ONPREM_OPTIONS,
    SAAS_OPTIONS,
)
from exasol.nb_connector.cli.param_wrappers import add_params
from exasol.nb_connector.cli.processing import processing


@cli.group()
def configure():
    """
    Add configuration options to the Secure Configuration Storage.
    """
    pass


@configure.command("onprem")
@add_params(ONPREM_OPTIONS)
def configure_onprem(scs_file: Path, **kwargs):
    """
    Configure connection to an Exasol on-premise instance.
    """
    processing.save(
        scs_file,
        StorageBackend.onprem,
        use_itde=False,
        values=kwargs,
    )


@configure.command("saas")
@add_params(SAAS_OPTIONS)
def configure_saas(scs_file: Path, **kwargs):
    """
    Configure connection to an Exasol SaaS instance.

    Configuring one of the parameters --saas-database-id and
    --saas-database-name is sufficient.
    """
    processing.save(
        scs_file,
        StorageBackend.saas,
        use_itde=False,
        values=kwargs,
    )


@configure.command("docker-db")
@add_params(DOCKER_DB_OPTIONS)
def configure_docker_db(scs_file: Path, **kwargs):
    """
    Configure connection to an Exasol Docker instance.
    """
    processing.save(
        scs_file,
        StorageBackend.onprem,
        use_itde=True,
        values=kwargs,
    )



import ipywidgets as widgets

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.connections import get_backend
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.common.ui_styles import config_styles


def select_db_backend(conf: Secrets) -> widgets.Widget:
    """
    Creates a UI form for choosing between the Exasol Docker-DB and an external Exasol Database.
    """

    ui_look = config_styles()

    db_options = ["Exasol Docker-DB", "Exasol On-Prem Database", "Exasol SaaS Database"]
    storage_backend = get_backend(conf)
    if storage_backend == StorageBackend.saas:
        db_choice = 2
    elif conf.get(CKey.use_itde, "True") == "True":
        db_choice = 0
    else:
        db_choice = 1
    db_selector = widgets.RadioButtons(
        options=db_options,
        value=db_options[db_choice],
        layout=ui_look.input_layout,
        style=ui_look.input_style,
    )
    select_btn = widgets.Button(
        description="Select", style=ui_look.button_style, layout=ui_look.button_layout
    )
    header_lbl = widgets.Label(
        value="Exasol Database Choice",
        style=ui_look.header_style,
        layout=ui_look.header_layout,
    )

    def select_database(btn):
        if db_selector.value == db_options[2]:
            conf.save(CKey.storage_backend, StorageBackend.saas.name)
        else:
            conf.save(CKey.storage_backend, StorageBackend.onprem.name)
            conf.save(CKey.use_itde, str(db_selector.value == db_options[0]))
        btn.icon = "check"

    def on_value_change(change):
        select_btn.icon = "pencil"

    select_btn.on_click(select_database)
    db_selector.observe(on_value_change, names=["value"])

    group_items = [header_lbl, widgets.Box([db_selector], layout=ui_look.row_layout)]
    items = [widgets.Box(group_items, layout=ui_look.group_layout), select_btn]
    ui = widgets.Box(items, layout=ui_look.outer_layout)
    return ui



# Initializing a JupySQL session
# This script is not intended to be run independently.
# It brings in the sql magic and performs common DB initialization steps,
# including opening the default schema and activating
# UDF languages at the session level.

from IPython import get_ipython
from IPython.core.error import UsageError

from exasol.nb_connector.connections import open_sqlalchemy_connection
from exasol.nb_connector.language_container_activation import get_activation_sql


def init(ai_lab_config):
    engine = open_sqlalchemy_connection(ai_lab_config)
    ipy = get_ipython()
    if ipy is None:
        raise RuntimeError(
            "Not running inside IPython. Magic commands will not execute."
        )
    ipy.run_line_magic("load_ext", "sql")
    ipy.run_line_magic("config", "SqlMagic.short_errors = False")
    ipy.push({"engine": engine}, interactive=True)
    try:
        ipy.run_line_magic("sql", "engine")
    except UsageError as e:
        if e.__cause__:
            raise e.__cause__
        else:
            raise e
    ipy.run_line_magic("sql", f"OPEN SCHEMA {ai_lab_config.db_schema}")
    activation_sql = get_activation_sql(ai_lab_config)
    ipy.run_line_magic("sql", activation_sql)



import click


def info(text: str):
    click.echo(click.style(text, fg="green"))


def success(text: str):
    click.echo(click.style(text, fg="green"))


def error(text: str | Exception):
    click.echo(click.style(f"Error: {text}", fg="bright_red"))


def warning(text: str):
    click.echo(click.style(f"Warning: {text}", fg="yellow"))



import pyexasol

_SETUP_SQL = [
    "OPEN SCHEMA {schema!i}",
    """
--/
            CREATE OR REPLACE JAVA SET SCRIPT IMPORT_PATH(...) EMITS (...) AS
              %scriptclass com.exasol.cloudetl.scriptclasses.FilesImportQueryGenerator;
              %jar {jar_path!r};
/
        """,
    """
--/
        CREATE OR REPLACE JAVA SCALAR SCRIPT IMPORT_METADATA(...) 
          EMITS (
                filename VARCHAR(2000), 
                partition_index VARCHAR(100), 
                start_index DECIMAL(36, 0), 
                end_index DECIMAL(36, 0)
          ) AS
          %scriptclass com.exasol.cloudetl.scriptclasses.FilesMetadataReader;
          %jar {jar_path!r};
/
        """,
    """
--/
        CREATE OR REPLACE JAVA SET SCRIPT IMPORT_FILES(...) EMITS (...) AS
          %scriptclass com.exasol.cloudetl.scriptclasses.FilesDataImporter;
          %jar {jar_path!r};
/
        """,
]


def setup_scripts(
    db_connection: pyexasol.ExaConnection, schema_name: str, bucketfs_jar_path: str
):
    """
    Perform initialization of scripts for could-storage-extension.
    :param db_connection: DB connection
    :param schema_name: name of the schema to be used.
    :param bucketfs_jar_path: path to cloud-storage-extension jar in BucketFS
    :return:
    """
    for sql in _SETUP_SQL:
        db_connection.execute(
            sql,
            query_params={
                "schema": schema_name,
                "jar_path": bucketfs_jar_path,
            },
        )



"""
Wrappers for adding custom properties to click parameters, e.g. SCS key.
"""

import getpass
import os
import re
from abc import abstractmethod
from typing import Any

import click

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli import reporting as report
from exasol.nb_connector.secret_store import Secrets


class ScsParam:
    """
    Abstract base class for ScsArgument and ScsOption.
    """

    def __init__(self, scs_key: CKey | None = None, **kwargs):
        self.scs_key = scs_key
        self._kwargs = dict(kwargs)

    def cli_option(self, full=False) -> str:
        return ""

    @property
    def arg_name(self) -> str:
        return ""

    def needs_entry(self, scs: Secrets) -> bool:
        return False

    @property
    def default(self) -> Any:
        return self._kwargs.get("default")

    def displayed_value(self, scs: Secrets) -> str | None:
        return None

    @abstractmethod
    def decorate(self, func):
        """
        This method is to be called when decorating the functions in the
        actual CLI declaration.
        """
        pass


class ScsArgument(ScsParam):
    """
    Represents a CLI argument for the SCS command.
    """

    def __init__(self, name: str, scs_key: CKey | None = None, **kwargs):
        super().__init__(scs_key, **kwargs)
        self.name = name

    @property
    def arg_name(self) -> str:
        return self.name

    def decorate(self, func):
        """
        This method is to be called when decorating the functions in the
        actual CLI declaration. Hence, ScsArgument calls click.argument()
        under the hood.
        """
        decorator = click.argument(self.name, **self._kwargs)
        return decorator(func)


class ScsOption(ScsParam):
    """
    CLI option for saving and checking values to the Secure Configuration
    Storage (SCS).

    In addition to the args supported by click.option() this class supports
    the following additional

    Parameters:
        scs_key:
            The related key in SCS or None if the option is not be stored in
            the SCS. ScsArgument For exaemple, ScsArgument scs_file is not to
            be stored in the SCS.

        scs_alternative_key:
            An alternative key for ScsOptions that are optional in case
            another option is provided. For example, for --saas-database-id
            you can specify --saas-database-name instead and vice-versa.

        scs_required:
            Whether this option is required to be stored in the SCS or only
            optional. This applies to --ssl-cert-path, for example.
    """

    def __init__(
        self,
        cli_option,
        *args,
        scs_key: CKey | None = None,
        scs_alternative_key: CKey | None = None,
        scs_required: bool = True,
        default: Any = None,
        help: str | None = None,
        **kwargs,
    ):
        _help = help if default is None else f"{help} [default: {default}]"
        super().__init__(scs_key=scs_key, help=_help, default=None, **kwargs)
        self._cli_option = cli_option
        self._args = args
        self.scs_alternative_key = scs_alternative_key
        self.scs_required = scs_required
        self._default = default

    def cli_option(self, full=False) -> str:
        raw = self._cli_option
        return raw if full else re.sub(r"/--.*$", "", raw)

    @property
    def default(self) -> Any:
        return self._default

    @property
    def arg_name(self) -> str:
        for arg in self._args:
            if not arg.startswith("--"):
                return arg
        name = self.cli_option()
        return name[2:].replace("-", "_")

    def decorate(self, func):
        """
        This method is to be called when decorating the functions in the
        actual CLI declaration. ScsOption calls click.option().
        """
        decorator = click.option(
            self._cli_option,
            *self._args,
            **self._kwargs,
            show_default=True,
        )
        return decorator(func)

    def displayed_value(self, scs: Secrets) -> str | None:
        return scs.get(self.scs_key) if self.scs_key else None

    def needs_entry(self, scs: Secrets) -> bool:
        """
        Return True, if the current option is configured to be saved to
        the SCS but SCS does not yet contain a value.
        """

        def has_value() -> bool:
            if not self.scs_key:
                return False
            if scs.get(self.scs_key) is not None:
                return True
            if alt := self.scs_alternative_key:
                return scs.get(alt) is not None
            return False

        return bool(self.scs_key) and self.scs_required and not has_value()

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}<{self.cli_option(full=True)}>"


class ScsSecretOption(ScsOption):
    """
    Represents a CLI option to be stored into SCS.
    """

    def __init__(
        self,
        name: str,
        envvar: str,
        prompt: str,
        scs_key: CKey,
        *args,
        metavar: str = "PASSWORD",
        **kwargs,
    ):
        super().__init__(
            name,
            *args,
            metavar=metavar,
            type=bool,
            is_flag=True,
            help=f"{prompt}  [env var: {envvar}]",
            scs_key=scs_key,
            **kwargs,
        )
        self.envvar = envvar
        self.prompt = prompt
        self.name = name

    def displayed_value(self, scs: Secrets) -> str | None:
        return "****" if self.scs_key and scs.get(self.scs_key) else None

    def get_secret(self, interactive: Any) -> str:
        """
        If interactive is True and the related environment variable is not
        set then ask for the secret interactively.
        """
        if value := os.getenv(self.envvar):
            report.info(f"Reading {self.name} from environment variable {self.envvar}.")
            return value
        if not interactive:
            return ""
        prompt = f"{self.prompt} (option {self.name}): "
        return getpass.getpass(prompt)


def add_params(scs_options: list[ScsArgument]):
    def multi_decorator(func):
        for o in reversed(scs_options):
            func = o.decorate(func)
        return func

    return multi_decorator



from pathlib import Path

from exasol.nb_connector.language_container_activation import ACTIVATION_KEY_PREFIX

DEFAULT_ALIAS = "ai_lab_default"  # To be used in the Jupyter notebooks of the AI Lab
PATH_IN_BUCKET = "container"

SLC_DOCKER_IMG_NAME = "exasol/script-language-container"

SLC_ACTIVATION_KEY_PREFIX = ACTIVATION_KEY_PREFIX + "slc_"
"""
Activation SQL for the Custom SLC will be saved in the Secure
Configuration Storage with this key.
"""

FLAVORS_PATH_IN_SLC_REPO = Path("flavors")
"""Path to flavors within the script-languages-release repository"""

SLC_RELEASE_TAG = "10.1.0"
"""
Using the SLC_RELEASE 10.1.0 because we are limited to slc-tool 3.*. (see pyproject.toml)
Check the developer guide (./doc/developer-guide.md) for more information.
"""

WORKSPACE_DIR = "slc_workspace"

SLC_GITHUB_REPO = "https://github.com/exasol/script-languages-release"
"""
GitHub URL which is used as base for the Script-Languages-Container.
"""

SLC_WORKSPACE_KEY_PREFIX = "slc_workspace_"
"""
Key prefix of the key in secret store to save the workspace directory. 
"""



"""Docker-related UI components.

This package exposes the public Docker-DB UI helpers.
"""

from .docker_db import (
    DockerDbDisplayStatus,
    docker_db_configuration,
    manage_docker_db,
)

__all__ = [
    "DockerDbDisplayStatus",
    "docker_db_configuration",
    "manage_docker_db",
]



"""Common UI utilities and components shared across notebook-connector UIs."""

from .jupysql import init
from .popup_message import display_popup
from .ui_styles import config_styles
from .useful_urls import Urls

__all__ = [
    "init",
    "display_popup",
    "config_styles",
    "Urls",
]



import sys
from pathlib import Path

from exasol.nb_connector.cli import reporting as report
from exasol.nb_connector.cli.groups import cli
from exasol.nb_connector.cli.options import SCS_OPTIONS
from exasol.nb_connector.cli.param_wrappers import add_params
from exasol.nb_connector.cli.processing import processing
from exasol.nb_connector.cli.processing.option_set import ScsCliError


@cli.command()
@add_params(SCS_OPTIONS)
def show(scs_file: Path):
    """
    Show the configuration currently saved to the Secure Configuration
    Storage.
    """
    try:
        processing.show_scs_content(scs_file)
    except ScsCliError as ex:
        report.error(ex)
        sys.exit(1)



from pathlib import Path

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import (
    ScsArgument,
    ScsOption,
)

SCS_OPTIONS = [
    ScsArgument(
        "scs_file",
        metavar="SCS_FILE",
        type=Path,
        required=True,
        envvar="SCS_FILE",
    ),
]

SAVE_OPTIONS = [
    ScsOption(
        "--overwrite-backend/--no-overwrite-backend",
        is_flag=True,
        help="Whether to overwrite a different backend in the SCS.",
    )
]

COMMON_OPTIONS = [
    ScsOption(
        "--db-schema",
        metavar="DB_SCHEMA",
        type=str,
        help="Database schema for installing UDFs of Exasol extensions",
        scs_key=CKey.db_schema,
    )
]



from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.extension_wrapper_common import (
    encapsulate_bucketfs_credentials,
)
from exasol.nb_connector.secret_store import Secrets

DEF_BFS_CONNECTION_NAME = "bfs_ai_lab_connection"

PATH_IN_BUCKET = "ai-lab"
""" Location in BucketFS bucket to upload data for Extensions, e.g. its language container. """


def ensure_bfs_connection_name(conf: Secrets) -> str:
    connection_name = conf.get(CKey.bfs_connection_name)
    if connection_name is None:
        conf.save(CKey.bfs_connection_name, DEF_BFS_CONNECTION_NAME)
        connection_name = DEF_BFS_CONNECTION_NAME
    return connection_name


def ensure_bfs_connection(conf: Secrets) -> None:
    """
    Creates a connection object in the database encapsulating
    a location in the BucketFS and BucketFS access credentials, if no connection was created yet.
    The path in the bucket is the hard-coded value for models.
    The connection name will be used from Secret-Store (Key='bfs_connection_name') if exists, otherwise
    "DEF_BFS_CONNECTION_NAME" will be used.
    Parameters:
         conf:
            The secret store.
    """
    encapsulate_bucketfs_credentials(
        conf,
        path_in_bucket=PATH_IN_BUCKET,
        connection_name=ensure_bfs_connection_name(conf),
    )



from importlib.metadata import version

from exasol_sagemaker_extension.deployment.deploy_create_statements import (
    DeployCreateStatements,
)
from exasol_sagemaker_extension.deployment.sme_language_container_deployer import (
    SmeLanguageContainerDeployer,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.connections import open_pyexasol_connection
from exasol.nb_connector.extension_wrapper_common import (
    deploy_language_container,
    encapsulate_aws_credentials,
)
from exasol.nb_connector.language_container_activation import (
    ACTIVATION_KEY_PREFIX,
    get_activation_sql,
)
from exasol.nb_connector.secret_store import Secrets

# Root directory in a BucketFS bucket where all stuff of the Sagemaker
# Extension, including its language container, will be uploaded.
PATH_IN_BUCKET = "SME"

LANGUAGE_ALIAS = "PYTHON3_SME"

LATEST_KNOWN_VERSION = version("exasol_sagemaker_extension")

# Activation SQL for the Sagemaker Extension will be saved in the secret
# store with this key.
ACTIVATION_KEY = ACTIVATION_KEY_PREFIX + "sme"

# Name of the connection object with AWS credentials and S3 location
# will be prefixed with this string.
AWS_CONNECTION_PREFIX = "SME_AWS"


def deploy_scripts(conf: Secrets) -> None:
    """
    Deploys all the extension's scripts to the database.

    Parameters:
        conf:
            The secret store. The store should contain the language activation
            SQL command. This command should be created during the language
            container deployment. The store should also have the DB schema.
    """

    with open_pyexasol_connection(conf, compression=True) as conn:
        # First need to activate the language container at the session level.
        activation_sql = get_activation_sql(conf)
        conn.execute(activation_sql)

        scripts_deployer = DeployCreateStatements(
            exasol_conn=conn,
            schema=conf.get(CKey.db_schema),
            to_print=False,
            develop=False,
        )
        scripts_deployer.run()


def initialize_sme_extension(
    conf: Secrets,
    version: str = LATEST_KNOWN_VERSION,
    language_alias: str = LANGUAGE_ALIAS,
    run_deploy_container: bool = True,
    run_deploy_scripts: bool = True,
    run_encapsulate_aws_credentials: bool = True,
    allow_override: bool = True,
) -> None:
    """
    Performs all necessary operations to get the Sagemaker Extension
    up and running. See the "Getting Started" and "Setup" sections of the
    extension's User Guide for details.

    Parameters:
        conf:
            The secret store. The store should contain all the required
            parameters for accessing the database, BucketFS and AWS.
        version:
            Sagemaker Extension version. If not specified the hardcoded
            latest known version will be used.
        language_alias:
            The language alias of the extension's language container. Normally
            this parameter would only be used for testing.
        run_deploy_container:
            If set to False will skip the language container deployment.
        run_deploy_scripts:
            If set to False will skip the deployment of the scripts.
        run_encapsulate_aws_credentials:
            If set to False will skip the creation of the database connection
            object encapsulating the AWS credentials.
        allow_override:
            If True allows overriding the language definition. Otherwise, if
            the database already has a language definition for the specified
            language alias, an attempt to deploy the container will result
            in a RuntimeError.
    """

    # Make the connection object name
    aws_conn_name = "_".join([AWS_CONNECTION_PREFIX, str(conf.get(CKey.db_user))])

    if run_deploy_container:
        container_url = SmeLanguageContainerDeployer.SLC_URL_FORMATTER.format(
            version=version
        )
        deploy_language_container(
            conf,
            container_url=container_url,
            container_name=SmeLanguageContainerDeployer.SLC_NAME,
            language_alias=language_alias,
            activation_key=ACTIVATION_KEY,
            path_in_bucket=PATH_IN_BUCKET,
            allow_override=allow_override,
        )

    # Create the required objects in the database
    if run_deploy_scripts:
        deploy_scripts(conf)
    if run_encapsulate_aws_credentials:
        encapsulate_aws_credentials(conf, aws_conn_name, CKey.sme_aws_bucket)

    # Save the connection object name in the secret store.
    conf.save(CKey.sme_aws_connection, aws_conn_name)



# pylint: skip-file
# Importing cython packages causes import errors in pylint, we need to investigate this later
# see https://github.com/exasol/notebook-connector/issues/206

import importlib.metadata
from pathlib import Path

from exasol.ai.text.deployment import license_deployment as txai_licenses
from exasol.ai.text.deployment.script_deployer import create_scripts
from exasol.ai.text.deployment.txaie_language_container_deployer import (
    TXAIELanguageContainerDeployer,
)
from exasol.ai.text.extraction.abstract_extraction import (
    AbstractExtraction,
    Defaults,
)
from exasol.ai.text.extraction.extraction import Extraction as TextAiExtraction
from exasol.ai.text.extractors.default_models import (
    DEFAULT_FEATURE_EXTRACTION_MODEL,
    DEFAULT_NAMED_ENTITY_MODEL,
    DEFAULT_NLI_MODEL,
)
from transformers import (
    AutoModel,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.bfs_connection import ensure_bfs_connection
from exasol.nb_connector.connections import (
    open_pyexasol_connection,
)
from exasol.nb_connector.extension_wrapper_common import (
    PATH_IN_BUCKET_FOR_SLC,
    deploy_language_container,
)
from exasol.nb_connector.language_container_activation import (
    ACTIVATION_KEY_PREFIX,
    get_activation_sql,
)
from exasol.nb_connector.model_installation import (
    TransformerModel,
    create_model_repository,
    ensure_model_subdir_config_value,
    install_model,
)
from exasol.nb_connector.secret_store import Secrets

# Models downloaded from the Huggingface archive to a local drive will be
# cached in directory MODELS_CACHE_DIR.
#
# TXAIE uses the same directories as TE (see function initialize_te_extension)
# as both extensions are using Huggingface Models. This also avoids confusion,
# and ensures backwards compatibility.
from exasol.nb_connector.transformers_extension_wrapper import (
    MODELS_CACHE_DIR,
)

LANGUAGE_ALIAS = "PYTHON3_TXAIE"

ACTIVATION_KEY = ACTIVATION_KEY_PREFIX + "txaie"
"""
Activation SQL for the Text AI Extension will be saved in the secret store
with this key.

TXAIE brings its own Script Language Container (SLC) which needs to be
activated by a dedicated SQL statement `ALTER SESSION SET SCRIPT_LANGUAGES`.  Applications
can store the language definition in the configuration store (SCS) from the Notebook
Connector's class `exasol.nb_connector.secret_store.Secrets`.

Using `ACTIVATION_KEY` as defined key, TXAIE can provide convenient interfaces
accepting only the SCS and retrieving all further data from the there.
"""

BFS_CONNECTION_PREFIX = "TXAIE_BFS"
"""
Prefix for Exasol CONNECTION objects containing a BucketFS location and
credentials.
"""


def deploy_license(
    conf: Secrets,
    license_content: str | None = None,
    license_file: Path | None = None,
) -> None:
    """
    Deploys the given license and saves its identifier to the secret store. The licence can either be
    defined by a path pointing to a licence file, or by the licence content given as a string.
    Parameters:
        conf:
            The secret store.
        licence_content:
            Optional. Content of a licence given as a string.
        licence_file:
            Optional. Path of a licence file.

    """

    license_content = license_content or txai_licenses.COMMUNITY_LICENSE
    if license_file:
        license_content = license_file.read_text()

    pyexasol_connection = open_pyexasol_connection(conf)
    txai_licenses.create_connection(pyexasol_connection, license_content)


def initialize_text_ai_extension(
    conf: Secrets,
    container_file: Path | None = None,
    version: str | None = None,
    install_slc: bool = True,
    install_scripts: bool = True,
    install_models: bool = True,
    allow_override_language_alias: bool = True,
) -> None:
    """
    Depending on which flags are set, runs different steps to install Text-AI Extension in the DB.
    Possible steps:

    * Call the Text-AI Extension's language container deployment API.

    If given a version, downloads the specified released version of the extension
    and uploads it to the BucketFS.

    If given a container_file path instead, installs the given container in the Bucketfs.

    If neither is given, attempts to install the latest version.

    This function doesn't activate the language container. Instead, it gets the
    activation SQL using the same API and writes it to the secret store. The name
    of the key is defined in the ACTIVATION_KEY constant.

    * Install default transformers models into
      the Bucketfs using Transformers Extensions upload model functionality.

    * Install Text-AI specific scripts.

    Parameters:
        conf:
            The secret store. The store must contain the DB connection parameters
            and the parameters of the BucketFS service.
        container_file:
            Optional. Path pointing to the locally stored Script Language Container file for the Text-AI Extension.
        version:
            Optional. Text-AI extension version.
        install_slc:
            If True runs deployment of the locally stored Script Language Container file for the Text-AI Extension.
        install_scripts:
            If True runs deployment of Text-AI Extension scripts.
        install_models:
            If True uploads default Transformers models to the BucketFS.
        install_bfs_credentials:
            If set to False will skip the creation of the text ai specific database connection
            object encapsulating the BucketFS credentials.
        allow_override_language_alias:
            If True allows overriding the language definition.
    """

    def deploy_slc(container_file=None, version=None):
        container_url = version and TXAIELanguageContainerDeployer.slc_download_url(
            version
        )
        deploy_language_container(
            conf,
            container_name=TXAIELanguageContainerDeployer.SLC_NAME,
            container_file=container_file,
            container_url=container_url,
            language_alias=LANGUAGE_ALIAS,
            activation_key=ACTIVATION_KEY,
            path_in_bucket=PATH_IN_BUCKET_FOR_SLC,
            allow_override=allow_override_language_alias,
        )

    print("Text AI: Updating Secure Configuration Storage")
    conf.save(CKey.txaie_models_cache_dir, MODELS_CACHE_DIR)

    print("Text AI: Ensuring BFS model connection")
    ensure_bfs_connection(conf)
    ensure_model_subdir_config_value(conf)

    if install_slc:
        print("Text AI: Downloading and installing Script Language Container (SLC)")
        if container_file:
            deploy_slc(container_file=container_file)
        else:
            version = version or importlib.metadata.version("exasol_text_ai_extension")
            deploy_slc(version=version)

    if install_models:
        #  Install default Hugging Face models into the Bucketfs using
        #  Transformers Extensions upload model functionality.
        print("Text AI: Downloading and installing Huggingface models to BucketFS:")
        install_model(
            conf,
            TransformerModel(
                DEFAULT_FEATURE_EXTRACTION_MODEL, "feature-extraction", AutoModel
            ),
        )
        install_model(
            conf,
            TransformerModel(
                DEFAULT_NAMED_ENTITY_MODEL,
                "token-classification",
                AutoModelForTokenClassification,
            ),
        )
        install_model(
            conf,
            TransformerModel(
                DEFAULT_NLI_MODEL,
                "zero-shot-classification",
                AutoModelForSequenceClassification,
            ),
        )

    if install_scripts:
        print("Text AI: Creating Scripts")
        pyexasol_connection = open_pyexasol_connection(
            conf, schema=conf.get(CKey.db_schema)
        )
        create_scripts(pyexasol_connection)

    print("Text AI: Installation finished.")


class Extraction(AbstractExtraction):
    def defaults_with_model_repository(self, conf: Secrets) -> Defaults:
        if self.defaults.model_repository:
            return self.defaults
        model_repository = create_model_repository(conf=conf)
        return Defaults(
            parallelism_per_node=self.defaults.parallelism_per_node,
            batch_size=self.defaults.batch_size,
            model_repository=model_repository,
        )

    def run(self, conf: Secrets) -> None:
        activation_sql = get_activation_sql(conf)
        defaults = self.defaults_with_model_repository(conf)
        with open_pyexasol_connection(conf, compression=True) as connection:
            connection.execute(query=activation_sql)
            TextAiExtraction(
                extractor=self.extractor,
                output=self.output,
                defaults=defaults,
            ).run(
                pyexasol_con=connection,
                temporary_db_object_schema=conf.db_schema,
                language_alias=LANGUAGE_ALIAS,
            )



from __future__ import annotations

import contextlib
import logging
import os
import shutil
from pathlib import Path

from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.slc import constants
from exasol.nb_connector.slc.git_access import GitAccess

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


@contextlib.contextmanager
def current_directory(dir: Path):
    """
    Changes the current working directory to the specified one and reverts
    to the previous one on exit.
    """
    previous = Path.cwd()
    os.chdir(dir)
    try:
        yield
    finally:
        os.chdir(previous)


class Workspace:
    def __init__(self, root_dir: Path):
        self.root_dir = root_dir

    @classmethod
    def for_slc(cls, name: str, secrets: Secrets) -> Workspace:
        workspace_key = constants.SLC_WORKSPACE_KEY_PREFIX + name
        path = Path.cwd()
        if path_from_secret_store := secrets.get(workspace_key):
            path = Path(path_from_secret_store)
        else:
            secrets.save(workspace_key, str(path))
        return cls(path / constants.WORKSPACE_DIR / name)

    def clone_slc_repo(self):
        """
        Clones the script-languages-release repository from Github into
        the target dir configured in the Secure Configuration Storage.
        """
        path = self.git_clone_path
        if path.is_dir():
            LOG.warning(f"Directory '{path}' is not empty. Checking consistency...")
            try:
                GitAccess.checkout_recursively(path)
                return
            except Exception as e:
                LOG.warning(
                    f"Git repository is inconsistent: {e}. Doing a fresh clone..."
                )
                shutil.rmtree(path)

        path.mkdir(parents=True, exist_ok=True)
        GitAccess.clone_from_recursively(
            constants.SLC_GITHUB_REPO, path, constants.SLC_RELEASE_TAG
        )

    @property
    def git_clone_path(self) -> Path:
        """
        Returns the path for cloning Git repository
        script-languages-container-release.
        """
        return self.root_dir / "git-clone"

    @property
    def export_path(self) -> Path:
        """
        Returns the export path for script-languages-container
        """
        return self.root_dir / "container"

    @property
    def output_path(self) -> Path:
        """
        Returns the output path containing caches and logs.
        """
        return self.root_dir / "output"

    def cleanup_output_path(self) -> None:
        """
        Remove the output path recursively.
        """
        shutil.rmtree(self.output_path)

    def cleanup_export_path(self) -> None:
        """
        Remove the export path recursively
        """
        shutil.rmtree(self.export_path)



from pathlib import Path

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import ScsOption

SSL_OPTIONS = [
    ScsOption(
        "--ssl-use-cert-validation/--no-ssl-use-cert-validation",
        type=bool,
        default=True,
        help="Whether to validate SSL certificates or not",
        scs_key=CKey.cert_vld,
    ),
    ScsOption(
        "--ssl-cert-path",
        metavar="FILE/DIR",
        type=Path,
        help="SSL trusted CA file/dir",
        scs_key=CKey.trusted_ca,
        scs_required=False,
    ),
]



import pyexasol

from exasol.nb_connector.connections import open_pyexasol_connection
from exasol.nb_connector.secret_store import Secrets

# All secret store entries with language container activation commands
# will have this common prefix in their keys.
ACTIVATION_KEY_PREFIX = "language_container_activation_"


def get_registered_languages_string(conf: Secrets) -> str:
    """
    Gets the session level language definitions from the database. Returns a string
    in a format of a space separated list: "<alias1>=<url1> <alias2=<url2> ..."
    """

    with open_pyexasol_connection(conf) as pyexasol_conn:
        query_result = pyexasol_conn.execute(
            "SELECT SESSION_VALUE FROM SYS.EXA_PARAMETERS WHERE PARAMETER_NAME='SCRIPT_LANGUAGES'"
        ).fetchall()
        return query_result[0][0]


def get_registered_languages(conf: Secrets) -> dict[str, str]:
    """
    Collects the existing, session level, language definitions from the database.
    Returns them as a dictionary {alias: language_url}
    """

    lang_definitions_str = get_registered_languages_string(conf)
    result: dict[str, str] = {}
    for lang_definition in lang_definitions_str.split():
        alias, lang_url = lang_definition.split("=", maxsplit=1)
        result[alias] = lang_url
    return result


def get_requested_languages(conf: Secrets) -> dict[str, str]:
    """
    Collects language definitions from the secret store. Returns them as a
    dictionary {alias: language_url}.

    Raises a RuntimeError if tt finds two entries with the same alias but
    different language definitions (i.e. URLs).
    """

    result: dict[str, str] = {}
    # Iterate over all entries that look like language definitions.
    for key, value in conf.items():
        if key.startswith(ACTIVATION_KEY_PREFIX):
            alias, lang_url = value.split("=", maxsplit=1)
            alias = alias.upper()
            if alias in result:
                if result[alias] != lang_url:
                    error = (
                        "Unable to merge multiple language definitions. "
                        f"Found incompatible definitions for the language alias {alias}."
                    )
                    raise RuntimeError(error)
            else:
                result[alias] = lang_url
    return result


def get_activation_sql(conf: Secrets) -> str:
    """
    Merges multiple language definitions (i.e. URLs) found in the secret store with
    language definitions currently registered in the database at the SESSION level.
    Returns a language activation command which combines all that language definitions.

    For details on how an activation command may look like please refer to
    https://docs.exasol.com/db/latest/database_concepts/udf_scripts/adding_new_packages_script_languages.htm

    The secret store entries containing language definitions should have keys with a
    common prefix. This prefix is defined in the ACTIVATION_KEY_PREFIX.

    Multiple language definitions in the secret store would normally have unique
    aliases. In case there are more than one entry with the same alias, their
    definitions should match. Otherwise, a RuntimeError will be raised. If a language
    with the same aliases is already present in the database, its definition will be
    overwritten, no exception raised.
    """

    # Collect language definitions, either already registered in the database or
    # recorded in the secret store.
    lang_definitions = get_registered_languages(conf)
    lang_definitions.update(get_requested_languages(conf))

    # Build and return an SQL command for the language container activation.
    merged_langs_str = " ".join(
        f"{key}={value}" for key, value in lang_definitions.items()
    )
    return f"ALTER SESSION SET SCRIPT_LANGUAGES='{merged_langs_str}';"


def open_pyexasol_connection_with_lang_definitions(
    conf: Secrets, **kwargs
) -> pyexasol.ExaConnection:
    """
    Opens a `pyexasol` connection and applies the `ALTER SESSION` command using all registered languages.
    """
    conn = open_pyexasol_connection(conf, **kwargs)
    conn.execute(get_activation_sql(conf))
    return conn



from __future__ import annotations

import json
from datetime import timedelta
from pathlib import Path

import exasol.bucketfs as bfs
from exasol.python_extension_common.deployment.extract_validator import ExtractValidator
from exasol.python_extension_common.deployment.language_container_deployer import (
    LanguageContainerDeployer,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.connections import (
    get_backend,
    get_external_host,
    get_saas_database_id,
    open_bucketfs_location,
    open_pyexasol_connection,
)
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.utils import optional_str_to_bool

PATH_IN_BUCKET_FOR_SLC = "ai-lab/slc"
"""
Location to deploy Script-Language-Contains relatively to bucket in BucketFS.
"""


def str_to_bool(conf: Secrets, key: CKey, default_value: bool) -> bool:
    """
    Tries to read a binary (i.e. yes/no) value from the secret store. If found
    returns the correspondent boolean. Otherwise, returns the provided default
    value.

    Parameters:
        conf:
            The secret store.
        key:
            The binary value key in the secret store.
        default_value:
            Default value if the key is not in the secret store.
    """
    prop_value = optional_str_to_bool(conf.get(key))
    return default_value if prop_value is None else prop_value


def _get_optional_external_host(conf: Secrets) -> str | None:
    """
    Get the host part of an onprem database URL if the data can be found
    in the configuration, otherwise None.
    """
    if conf.get(CKey.db_host_name) and conf.get(CKey.db_port):
        return get_external_host(conf)
    return None


def _get_optional_bfs_port(conf: Secrets) -> int | None:
    """
    Return the BucketFS service port number if it can be found in the
    configuration, otherwise None.
    """
    port_str = conf.get(CKey.bfs_port)
    if port_str:
        return int(port_str)
    return None


def deploy_language_container(
    conf: Secrets,
    path_in_bucket: str,
    language_alias: str,
    activation_key: str,
    container_url: str | None = None,
    container_file: Path | None = None,
    container_name: str | None = None,
    allow_override: bool = True,
    timeout: timedelta = timedelta(minutes=10),
) -> None:
    """
    Downloads language container from the specified location and uploads it to the
    BucketFS.

    This function doesn't activate the language container. Instead, it generates the
    activation SQL command and writes it to the secret store using the provided key.

    Parameters:
        conf:
            The secret store. The store must contain the DB connection parameters
            and the parameters of the BucketFS service.
        path_in_bucket:
            Path in the BucketFS where the container should be saved.
        language_alias:
            The language alias of the extension's language container.
        activation_key:
            A secret store key for saving the activation SQL.
        container_url:
            An optional URL to download the language container from.
            Either the `container_url` or `container_file` must be provided,
            otherwise a ValueError will be raised.
        container_file:
            An optional path of the container file (*.tar.gz) in a local file system.
            Either the `container_url` or `container_file` must be provided,
            otherwise a ValueError will be raised.
        container_name:
            If provided, the language container will be saved in given bucket of
            BucketFS with this filename. Otherwise, the name of the container file
            will be used.
        allow_override:
            If True allows overriding the language definition.
        timeout:
            Maximum time allowed for saving the container in the BucketFS.
            This includes the time required to unpack the container.
            The downloading time is not included.
    """

    with open_pyexasol_connection(conf, compression=True) as conn:
        validator = ExtractValidator(conn, timeout)
        bucketfs_location = open_bucketfs_location(conf) / path_in_bucket
        deployer = LanguageContainerDeployer(
            pyexasol_connection=conn,
            language_alias=language_alias,
            bucketfs_path=bucketfs_location,
            extract_validator=validator,
        )

        if container_file:
            deployer.run(
                container_file,
                container_name,
                alter_system=False,
                allow_override=allow_override,
                wait_for_completion=True,
            )
        elif container_url:
            deployer.download_and_run(
                container_url,
                container_name,
                alter_system=False,
                allow_override=allow_override,
                wait_for_completion=True,
            )
        else:
            raise ValueError("Either container URL or container file must be provided")

        # Install the language container.
        # Save the activation SQL in the secret store.
        language_def = deployer.get_language_definition(container_name)
        conf.save(activation_key, language_def)


def encapsulate_bucketfs_credentials(
    conf: Secrets, path_in_bucket: str, connection_name: str
) -> None:
    """
    Creates a connection object in the database encapsulating
    a location in the BucketFS and BucketFS access credentials.

    Parameters:
        conf:
            The secret store.
            For an On-Prem database the store must hold the BucketFS service
            parameters (bfs_host_name or db_host_name, bfs_port,
            bfs_service), the access credentials (bfs_user,
            bfs_password), and the bucket name (bfs_bucket), as well
            as the DB connection parameters.
            For a SaaS database the store must hold the SaaS connection
            parameters (saas_url, saas_account_id, saas_database_id or
            saas_database_name, saas_token).
        path_in_bucket:
            Path identifying a location in the bucket.
        connection_name:
            Name for the connection object to be created.

    The parameters will be stored in json strings. The distribution
    of the parameters among the connection entities will be as following.
    On-Prem:
        TO: backend, url, service_name, bucket_name, verify, path
        USER: username
        IDENTIFIED BY: password
    SaaS:
        TO: backend, url, account_id, path
        USER: database_id
        IDENTIFIED BY: pat

    Note that the parameter names correspond to the arguments of the build_path
    function. provided by the bucketfs-python. Hence, the most convenient way to
    handle this lot is to combine json structures from all three entities and
    pass them as kwargs to the build_path. The code below gives a usage example.

    bfs_params = json.loads(conn_obj.address)
    bfs_params.update(json.loads(conn_obj.user))
    bfs_params.update(json.loads(conn_obj.password))
    bfs_path = build_path(**bfs_params)

    A note about handling the TLS certificate verification settings.
    If the server certificate verification is turned on, either through
    reliance of the default https request settings or by setting the cert_vld
    configuration parameter to True, this intention will be passed to
    the connection object. However, if the user specifies a custom CA list
    file or directory, which also implies the certificate verification,
    the connection object will instead turn the verification off. This is
    because there is no guarantee that the consumer of the connection object,
    i.e. a UDF, would have this custom CA list, and even if it would, its location
    is unknown. This is only applicable for an On-Prem backend.
    """

    def to_json_str(**kwargs) -> str:
        filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return json.dumps(filtered_kwargs)

    backend = get_backend(conf)
    if backend == StorageBackend.onprem:
        # Here we are using the internal bucket-fs host and port, falling back
        # to the external parameters if the former are not specified.
        host = conf.get(
            CKey.bfs_internal_host_name,
            conf.get(CKey.bfs_host_name, conf.get(CKey.db_host_name)),
        )
        port = conf.get(CKey.bfs_internal_port, conf.get(CKey.bfs_port))
        protocol = "https" if str_to_bool(conf, CKey.bfs_encryption, True) else "http"
        url = f"{protocol}://{host}:{port}"
        verify: bool | None = (
            False
            if conf.get(CKey.trusted_ca)
            else optional_str_to_bool(conf.get(CKey.cert_vld))
        )
        conn_to = to_json_str(
            backend=bfs.path.StorageBackend.onprem.name,
            url=url,
            service_name=conf.get(CKey.bfs_service),
            bucket_name=conf.get(CKey.bfs_bucket),
            path=path_in_bucket,
            verify=verify,
        )
        conn_user = to_json_str(username=conf.get(CKey.bfs_user))
        conn_password = to_json_str(password=conf.get(CKey.bfs_password))
    else:
        database_id = get_saas_database_id(conf)
        conn_to = to_json_str(
            backend=bfs.path.StorageBackend.saas.name,
            url=conf.get(CKey.saas_url),
            account_id=conf.get(CKey.saas_account_id),
            path=path_in_bucket,
        )
        conn_user = to_json_str(database_id=database_id)
        conn_password = to_json_str(pat=conf.get(CKey.saas_token))

    sql = f"""
    CREATE OR REPLACE CONNECTION [{connection_name}]
        TO {{BUCKETFS_ADDRESS!s}}
        USER {{BUCKETFS_USER!s}}
        IDENTIFIED BY {{BUCKETFS_PASSWORD!s}}
    """
    query_params = {
        "BUCKETFS_ADDRESS": conn_to,
        "BUCKETFS_USER": conn_user,
        "BUCKETFS_PASSWORD": conn_password,
    }
    with open_pyexasol_connection(conf, compression=True) as conn:
        conn.execute(query=sql, query_params=query_params)


def encapsulate_huggingface_token(conf: Secrets, connection_name: str) -> None:
    """
    Creates a connection object in the database encapsulating a Huggingface token.

    Parameters:
        conf:
            The secret store. The store must hold the Huggingface token (huggingface_token),
             as well as the DB connection parameters.
        connection_name:
            Name for the connection object to be created.
    """

    sql = f"""
    CREATE OR REPLACE CONNECTION [{connection_name}]
        TO ''
        IDENTIFIED BY {{TOKEN!s}}
    """
    query_params = {"TOKEN": conf.get(CKey.huggingface_token)}
    with open_pyexasol_connection(conf, compression=True) as conn:
        conn.execute(query=sql, query_params=query_params)


def encapsulate_aws_credentials(
    conf: Secrets, connection_name: str, s3_bucket_key: CKey
) -> None:
    """
    Creates a connection object in the database encapsulating the address of
    an AWS S3 bucket and AWS access credentials.

    Parameters:
        conf:
            The secret store. The store must hold the S3 bucket parameters
            (aws_bucket, aws_region) and AWS access credentials (aws_access_key_id,
            aws_secret_access_key), as well as the DB connection parameters.
        connection_name:
            Name for the connection object to be created.
        s3_bucket_key:
            The secret store key of the AWS S3 bucket name.
    """

    sql = f"""
    CREATE OR REPLACE  CONNECTION [{connection_name}]
        TO 'https://{conf.get(s3_bucket_key)}.s3.{conf.get(CKey.aws_region)}.amazonaws.com/'
        USER {{ACCESS_ID!s}}
        IDENTIFIED BY {{SECRET_KEY!s}}
    """
    query_params = {
        "ACCESS_ID": conf.get(CKey.aws_access_key_id),
        "SECRET_KEY": conf.get(CKey.aws_secret_access_key),
    }
    with open_pyexasol_connection(conf, compression=True) as conn:
        conn.execute(query=sql, query_params=query_params)



from typing import (
    Any,
)

import docker
import ifaddr
from docker.models.containers import Container


class IPRetriever:
    def ips(self) -> list[ifaddr.IP]:
        adapters = ifaddr.get_adapters()
        ips = [ip for adapter in adapters for ip in adapter.ips]
        return ips


def retrieve_networks_of_container(
    container: Container,
) -> list[dict[str, Any]]:
    container.reload()
    network_settings = container.attrs["NetworkSettings"]
    networks = network_settings["Networks"].values()
    return networks


class ContainerByIp:
    """
    Find a Docker container by ip addresses of its networks
    being included in the given list of ip addresses.
    A single matching ip address is sufficient.
    """

    def __init__(self, docker_client: docker.DockerClient):
        self._docker_client = docker_client

    def find(self, ip_addresses: list[str]) -> Container | None:
        candidates = {
            container
            for container in self._docker_client.containers.list()
            for network in retrieve_networks_of_container(container)
            if network["IPAddress"] in ip_addresses
        }
        if len(candidates) == 1:
            return next(iter(candidates))
        elif len(candidates) == 0:
            return None
        else:

            def format(container):
                return f'{container.id[:12]} "{container.name}"'

            pretty = ", ".join(format(c) for c in candidates)
            raise RuntimeError(f"Found multiple matching containers: {pretty}")



from itertools import chain

import ipywidgets as widgets

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.common.ui_styles import config_styles


def get_config(
    secrets: Secrets,
    inputs: list[list[tuple[str, widgets.Widget, CKey]]],
    group_names: list[str],
) -> widgets.Widget:
    """
    The function creates a generic configuration editor UI.
    The UI consists of one or more blocks of configuration data aligned vertically.
    Each block has a header and one or more rows with input fields. Each row has a
    label on the left and one input widget on the right. The form ends with a "Save"
    button, clicking on which results in saving the data in the configuration store.

    Parameters
        conf:        Configuration store
        inputs:      List of the input blocks. Each block is a list of input rows.
                     Each row consists of the label text, input widget, and the key
                     of the corresponding configuration element in the store.
        group_names: Header names for the blocks. The length of this list should
                     match the length of the inputs.
    """

    ui_look = config_styles()
    save_btn = widgets.Button(
        description="Save", style=ui_look.button_style, layout=ui_look.button_layout
    )

    def save(btn):
        for row in chain(*inputs):
            _, widget, key = row
            secrets.save(key, str(widget.value))
        btn.icon = "check"

    def on_value_change(change):
        # Icon "pen" in font ... is not available on all systems
        save_btn.icon = "pencil"

    save_btn.on_click(save)

    # Apply the styles and layouts to the input fields
    for row in chain(*inputs):
        widget = row[1]
        widget.style = ui_look.input_style
        widget.layout = ui_look.input_layout
        widget.observe(on_value_change, names=["value"])
    # Create a list of lists with input rows
    item_groups = [
        [
            widgets.Box(
                [
                    widgets.Label(
                        value=input_title,
                        style=ui_look.label_style,
                        layout=ui_look.label_layout,
                    ),
                    input_widget,
                ],
                layout=ui_look.row_layout,
            )
            for input_title, input_widget, _ in input_group
        ]
        for input_group in inputs
    ]
    # Create a list of blocks
    items = [
        widgets.Box(
            [
                widgets.Label(
                    value=group_name,
                    style=ui_look.header_style,
                    layout=ui_look.header_layout,
                )
            ]
            + item_group,
            layout=ui_look.group_layout,
        )
        for item_group, group_name in zip(item_groups, group_names)
    ]
    # Add the save button and put everything in an outer Box.
    items.append(save_btn)
    ui = widgets.Box(items, layout=ui_look.outer_layout)
    return ui



"""
Bucketfs-related functions.
"""

import logging
import pathlib

import exasol.bucketfs as bfs

_logger = logging.getLogger(__name__)


def _file_in_bucket(file_name: str, bucket: bfs.Bucket) -> bool:
    """
    Checks that given file name is present in bucket.
    :param file_name: name to check
    :param bucket: bucket object
    :return: True if name is present, else False
    """
    try:
        return file_name in list(bucket)
    except TypeError as e:
        # SaaSBucket is not iterable, need to handle it here
        if e.args[0].endswith("not iterable"):
            return False
        raise e


def put_file(
    bucket: bfs.Bucket, file_path: pathlib.Path, skip_if_exists: bool = True
) -> bfs.path.BucketPath:
    """
    Uploads given file into bucketfs
    :param bucket: bucket to use
    :param file_path: local file path to uplaod. File have to exist.
    :param skip_if_exists: Do not upload if file already present in the bucketfs.
    :return: Path in the bucketfs.
    """
    if not file_path.exists():
        raise ValueError(f"Local file doesn't exist: {file_path}")
    local_name = file_path.name
    if skip_if_exists and _file_in_bucket(local_name, bucket):
        _logger.info("File %s is already present in the bucketfs", local_name)
    else:
        _logger.info("Uploading file %s to bucketfs", local_name)
        with file_path.open("rb") as file:
            bucket.upload(local_name, file)
    return bfs.path.BucketPath(local_name, bucket)



from pathlib import Path


def upward_file_search(file_name: str) -> str:
    """
    Looks for a specified file starting from the current directory upward the file hierarchy.
    Hence, the last place to find the file is the root.

    Returns the full path of the file if found.
    Otherwise, raises a ValueError exception.
    """

    dir = Path().resolve()
    while dir.name:
        maybe_file = dir / file_name
        if maybe_file.is_file():
            return str(maybe_file)
        dir = dir.parent
    raise ValueError(f"Cannot find {file_name}")


def optional_str_to_bool(value: str | None) -> bool | None:
    """
    Converts an optional string value to an optional boolean.
    None, '' => None.
    Case-insensitive "y", "yes", "true" => True
    Case-insensitive "n", "no", "false" => False
    Any other value cause a ValueError exception.
    """
    if not value:
        return None

    mapping = {
        "y": True,
        "yes": True,
        "true": True,
        "n": False,
        "no": False,
        "false": False,
    }
    key = value.lower()

    try:
        result = mapping[key]
    except KeyError as ex:
        raise ValueError("Invalid boolean value " + value) from ex

    return result



from __future__ import annotations

import getpass
import os
from pathlib import Path
from typing import Any

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.ai_lab_config import StorageBackend
from exasol.nb_connector.cli import reporting as report
from exasol.nb_connector.cli.options import (
    DOCKER_DB_OPTIONS,
    ONPREM_OPTIONS,
    SAAS_OPTIONS,
)
from exasol.nb_connector.cli.param_wrappers import (
    ScsOption,
    ScsParam,
)
from exasol.nb_connector.cli.processing.backend_selector import BackendSelector
from exasol.nb_connector.secret_store import Secrets


class ScsCliError(Exception):
    """
    Indicates an error when saving or checking CLI options wrt. the Secure
    Configuration Storage (SCS).
    """


SELECT_BACKEND_OPTION = ScsOption("backend", scs_key=CKey.storage_backend)
USE_ITDE_OPTION = ScsOption("use_itde", scs_key=CKey.use_itde)


def get_options(
    backend: StorageBackend,
    use_itde: bool,
) -> list[ScsParam]:
    def specific_options():
        if backend == StorageBackend.saas:
            return SAAS_OPTIONS
        if backend != StorageBackend.onprem:
            raise ScsCliError(f"Unsupported backend {backend}")
        return DOCKER_DB_OPTIONS if use_itde else ONPREM_OPTIONS

    return [SELECT_BACKEND_OPTION, USE_ITDE_OPTION] + specific_options()


class OptionSet:
    """
    * Find an option by its underscored name.

    * Transfer option values as default to other options.

    * Check the SCS content for completeness.
    """

    def __init__(self, scs: Secrets, backend: StorageBackend, use_itde: bool):
        self.scs = scs
        self.options = get_options(backend, use_itde)

    def default_values(self, values: dict[str, Any]) -> dict[str, Any]:
        """
        Return a dict of option arg names and default values.

        The dict contains only entries for options with a defined default
        value, that are neither specified explicitly, nor stored in the SCS,
        yet.

        For details, see https://github.com/exasol/notebook-connector/issues/285.
        """

        def use_default(option: ScsParam) -> bool:
            return bool(
                option.default is not None
                and values.get(option.arg_name) is None
                and option.scs_key
                and self.scs.get(option.scs_key) is None
            )

        return {o.arg_name: o.default for o in self.options if use_default(o)}

    def find_option(self, arg_name: str) -> ScsParam:
        """
        Find the full definition of a click parameter for the specified
        arg name.
        """
        try:
            return next(o for o in self.options if o.arg_name == arg_name)
        except StopIteration:
            raise ScsCliError(
                f"Couldn't find any option with parameter name {arg_name}."
            )

    def check(self):
        """
        Check if the content of the SCS is complete wrt. the selected
        backend as the required options depend on the selected backend.
        """
        missing = [
            o.cli_option(full=True) for o in self.options if o.needs_entry(self.scs)
        ]
        if not missing:
            config = BackendSelector(self.scs)
            report.success(
                "Configuration is complete for an "
                f"Exasol {config.backend_name} instance."
            )
            return
        formatted = ", ".join(missing)
        n = len(missing)
        prefix = "1 option is" if n == 1 else f"{n} options are"
        raise ScsCliError(f"{prefix} not yet configured: {formatted}.")


def get_scs_master_password():
    """
    Retrieve the master password for the SCS either from the related
    environment variable or by asking the user to type the password
    interactively.
    """
    if from_env := os.getenv("SCS_MASTER_PASSWORD"):
        return from_env
    return getpass.getpass("SCS master password: ")


def get_scs(scs_file: Path) -> Secrets:
    scs_password = get_scs_master_password()
    return Secrets(scs_file, scs_password)


def get_option_set(scs_file: Path) -> OptionSet:
    """
    Return an instance of an OptionSet if the SCS contains a proper
    backend selection. Otherwise raise an ScsCliError.

    This function is designed to be called only once in the CLI application.
    Otherwise it will always ask for the SCS master password and potentially
    report errors in case of no backend being selected.

    If you want to instantiate the OptionSet more often, then simply use
    its constructor.
    """

    scs = get_scs(scs_file)
    config = BackendSelector(scs)
    if not config.knows_backend:
        raise ScsCliError(f"SCS {scs_file} does not contain any backend.")
    return OptionSet(scs, config.backend, config.use_itde)



import logging
from pathlib import Path

from exasol.nb_connector.slc.package_types import (
    CondaPackageDefinition,
    PipPackageDefinition,
)
from exasol.nb_connector.slc.slc_error import SlcError


def _read_packages(file_path: Path, package_definition: type) -> list:
    packages = []
    with file_path.open("r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue  # skip empty or commented lines
            # Take everything before the '|' as package name
            package, version = line.split("|", 1)
            packages.append(package_definition(package, version))
    return packages


def _ends_with_newline(file_path: Path) -> bool:
    content = file_path.read_text()
    return content.endswith("\n")


def _filter_packages(
    original_packages: list[PipPackageDefinition] | list[CondaPackageDefinition],
    new_packages: list[PipPackageDefinition] | list[CondaPackageDefinition],
) -> list[PipPackageDefinition | CondaPackageDefinition]:
    filtered_packages = []
    for package in new_packages:
        add_package = True
        for original_package in original_packages:
            if package.pkg == original_package.pkg:
                add_package = False
                if package.version == original_package.version:
                    logging.warning("Package already exists: %s", original_package)
                else:
                    raise SlcError(
                        "Package already exists: %s but with different version",
                        original_package,
                    )
        if add_package:
            filtered_packages.append(package)
    return filtered_packages


def append_packages(
    file_path: Path,
    package_definition: type,
    packages: list[PipPackageDefinition] | list[CondaPackageDefinition],
):
    """
    Appends packages to the custom packages file.
    """
    original_packages = _read_packages(file_path, package_definition)
    filtered_packages = _filter_packages(original_packages, packages)
    ends_with_newline = _ends_with_newline(file_path)
    if filtered_packages:
        with open(file_path, "a") as f:
            if not ends_with_newline:
                f.write("\n")
            for p in filtered_packages:
                print(f"{p.pkg}|{p.version}", file=f)



from __future__ import annotations

from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.slc.slc_error import SlcError


class SlcFlavor:
    def __init__(self, slc_name: str):
        self.slc_name = slc_name

    @property
    def key(self):
        return f"SLC_FLAVOR_{self.slc_name.upper()}"

    def save(self, secrets: Secrets, flavor: str) -> None:
        secrets.save(self.key, flavor)

    def exists(self, secrets: Secrets) -> bool:
        return True if secrets.get(self.key) else False

    def verify(self, secrets: Secrets) -> str:
        try:
            return secrets[self.key]
        except AttributeError as ex:
            raise SlcError(
                "Secure Configuration Storage does not contain a"
                f" flavor for SLC {self.slc_name}."
            ) from ex



from enum import (
    Enum,
    auto,
)


class AILabConfig(Enum):
    use_itde = auto()
    db_host_name = auto()
    db_port = auto()
    db_schema = auto()
    db_user = auto()
    db_password = auto()
    db_encryption = auto()
    cert_vld = auto()
    trusted_ca = auto()
    client_cert = auto()
    client_key = auto()
    bfs_host_name = auto()
    bfs_port = auto()
    bfs_internal_host_name = auto()
    bfs_internal_port = auto()
    bfs_service = auto()
    bfs_bucket = auto()
    bfs_user = auto()
    bfs_password = auto()
    bfs_encryption = auto()
    mem_size = auto()
    disk_size = auto()
    huggingface_token = auto()
    aws_region = auto()
    aws_access_key_id = auto()
    aws_secret_access_key = auto()
    itde_container = auto()
    itde_volume = auto()
    itde_network = auto()
    te_hf_connection = auto()
    te_models_cache_dir = auto()
    txaie_models_cache_dir = auto()
    sme_aws_bucket = auto()
    sme_aws_role = auto()
    sme_aws_connection = auto()
    saas_url = auto()
    saas_token = auto()
    saas_account_id = auto()
    saas_database_id = auto()
    saas_database_name = auto()
    storage_backend = auto()
    accelerator = auto()
    bfs_model_subdir = auto()
    bfs_connection_name = auto()


class StorageBackend(Enum):
    onprem = auto()
    saas = auto()


class Accelerator(Enum):
    none = "none"
    nvidia = "nvidia"



from __future__ import annotations

import contextlib
import logging
import threading
from collections.abc import Iterable
from inspect import cleandoc
from pathlib import Path
from typing import (
    Any,
)

import tenacity
from sqlcipher3 import dbapi2 as sqlcipher  # type: ignore
from tenacity import (
    retry_if_exception_message,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey

LOG = logging.getLogger(__name__)
TABLE_NAME = "secrets"


class InvalidPassword(Exception):
    """Signal potentially incorrect master password."""


class Secrets:
    def __init__(self, db_file: Path, master_password: str) -> None:
        self.db_file = db_file
        self._master_password = master_password
        self._lock = threading.Lock()
        self._cache: dict[int, sqlcipher.Connection] = {}
        self._initialize()

    def _initialize(self) -> None:
        if self.db_file.exists():
            self._verify_access()
            return
        LOG.info('Creating file %s and table "%s"', self.db_file, TABLE_NAME)
        self._execute(f"CREATE TABLE {TABLE_NAME} (key TEXT PRIMARY KEY, value TEXT)")

    def connection(self) -> sqlcipher.Connection:
        """
        SQLite allows a connection to be used only by a single thread.  In
        multi-threaded scenarios we therefore need to maintain a connection
        pool, containing a separate connection for each thread.

        Potential exceptions:
        sqlcipher3.dbapi2.IntegrityError: UNIQUE constraint failed: secrets.key
        sqlcipher3.dbapi2.OperationalError: database is locked
        """
        thread_id = threading.get_ident()
        with self._lock:
            if con := self._cache.get(thread_id):
                return con
            con = sqlcipher.connect(self.db_file)  # pylint: disable=E1101
            self._cache[thread_id] = con
        with self._cursor(con) as cur:
            self._use_master_password(cur)
        return con

    def close(self) -> None:
        thread_id = threading.get_ident()
        with self._lock:
            con = self._cache.pop(thread_id, None)
        if con:
            con.close()

    def close_all(self) -> None:
        """
        Close all connections in cache and empty cache.
        """
        with self._lock:
            for con in self._cache.values():
                con.close()
            self._cache = {}

    def __del__(self) -> None:
        self.close()

    def _use_master_password(self, cur: sqlcipher.Cursor) -> None:
        """
        If database is unencrypted then this method encrypts it.
        If database is already encrypted then this method enables to access the data.
        """
        if self._master_password is not None:
            sanitized = self._master_password.replace("'", "\\'")
            cur.execute(f"PRAGMA key = '{sanitized}'")

    def _verify_access(self) -> None:
        try:
            self._execute("SELECT * FROM sqlite_master")
        # fmt: off
        except (sqlcipher.DatabaseError) as ex:  # pylint: disable=E1101
            # fmt: on
            LOG.error("Exception %s", ex)
            if str(ex) == "file is not a database":
                raise InvalidPassword(
                    cleandoc(
                        f"""
                    Cannot access
                    database file {self.db_file}.
                    This also happens if master password is incorrect.
                    """
                    )
                ) from ex
            raise

    # If the database is locked, wait exponentially min. 0.1 seconds, max. 5
    # seconds (50 * 0.1) and retry executing the current SQL statement.
    @tenacity.retry(
        retry=(
            retry_if_exception_type(sqlcipher.OperationalError)
            & retry_if_exception_message("database is locked")
        ),
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=0.1, min=0.1, max=30),
    )
    def _execute(
        self, stmt: str, args: list[Any] | None = None, cur: sqlcipher.Cursor = None
    ) -> None:
        """
        cur.execute() returns the same object on which execute() was
        called. To avoid lifetime issues of the cursor object, we dont't
        return it. This forces the caller to inject the cursor with the correct
        lifetime into this method.
        """
        with contextlib.ExitStack() as stack:
            cur = cur or stack.enter_context(self._cursor())
            cur.execute(stmt, args or [])

    @contextlib.contextmanager
    def _cursor(
        self,
        con: sqlcipher.Connection | None = None,
    ) -> sqlcipher.Cursor:
        con = con or self.connection()
        cur = con.cursor()
        try:
            yield cur
            con.commit()
        except:
            con.rollback()
            raise
        finally:
            cur.close()

    def save(self, key: str | CKey, value: str) -> Secrets:
        """key represents a system, service, or application"""
        key = key.name if isinstance(key, CKey) else key
        stmt = (
            f"INSERT INTO {TABLE_NAME} (key,value) VALUES (?, ?)"
            " ON CONFLICT(key) DO UPDATE SET value=?"
        )
        self._execute(stmt, [key, value, value])
        return self

    def get(self, key: str | CKey, default_value: str | None = None) -> str | None:
        key = key.name if isinstance(key, CKey) else key
        with self._cursor() as cur:
            self._execute(f"SELECT value FROM {TABLE_NAME} WHERE key=?", [key], cur=cur)
            row = cur.fetchone()
        return row[0] if row else default_value

    def __getattr__(self, key) -> str:
        val = self.get(key)
        if val is None:
            raise AttributeError(f'Unknown key "{key}"')
        return val

    def __getitem__(self, item) -> str:
        val = self.get(item)
        if val is None:
            raise AttributeError(f'Unknown key "{item}"')
        return val

    def keys(self) -> Iterable[str]:
        """Iterator over keys akin to dict.keys()"""
        with self._cursor() as cur:
            self._execute(f"SELECT key FROM {TABLE_NAME}", cur=cur)
            yield from (row[0] for row in cur)

    def values(self) -> Iterable[str]:
        """Iterator over values akin to dict.values()"""
        with self._cursor() as cur:
            self._execute(f"SELECT value FROM {TABLE_NAME}", cur=cur)
            yield from (row[0] for row in cur)

    def items(self) -> Iterable[tuple[str, str]]:
        """Iterator over keys and values akin to dict.items()"""
        with self._cursor() as cur:
            self._execute(f"SELECT key, value FROM {TABLE_NAME}", cur=cur)
            yield from ((row[0], row[1]) for row in cur)

    def remove(self, key: str | CKey) -> None:
        """
        Deletes the entry with the specified key if it exists.  Doesn't
        raise any exception if the key doesn't exist.
        """
        key = key.name if isinstance(key, CKey) else key
        self._execute(f"DELETE FROM {TABLE_NAME} WHERE key=?", [key])



import logging
import os
from enum import IntFlag

import docker
from docker.models.networks import Network
from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    LATEST_DB_VERSION,
)
from exasol_integration_test_docker_environment.lib import api
from exasol_integration_test_docker_environment.lib.docker import (
    ContextDockerClient,
)
from exasol_integration_test_docker_environment.lib.docker.container.utils import (
    remove_docker_container,
)
from exasol_integration_test_docker_environment.lib.docker.networks.utils import (
    remove_docker_networks,
)
from exasol_integration_test_docker_environment.lib.docker.volumes.utils import (
    remove_docker_volumes,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)

from exasol.nb_connector.ai_lab_config import (
    Accelerator,
    AILabConfig,
)
from exasol.nb_connector.container_by_ip import (
    ContainerByIp,
    IPRetriever,
)
from exasol.nb_connector.secret_store import Secrets

ENVIRONMENT_NAME = "DemoDb"
NAME_SERVER_ADDRESS = "8.8.8.8"
TEST_DB_VERSION_ENV_VAR = "TEST_DB_VERSION"


class ItdeContainerStatus(IntFlag):
    ABSENT = 0
    STOPPED = 1
    RUNNING = 3
    VISIBLE = 5
    READY = RUNNING | VISIBLE


def bring_itde_up(conf: Secrets, env_info: EnvironmentInfo | None = None) -> None:
    """
    Launches the ITDE environment using its API. Sets hardcoded environment name,
    and Google name server address. Additionally, can set the following
    parameters with values collected from the secret store:
    - database port forwarding,
    - BucketFS port forwarding,
    - database memory size (the value is assumed to be the number of gigabytes),
    - database disk size (the value is assumed to be the number of gigabytes).

    Optionally, an existing instance of the DockerDB can be used instead. In this
    case the EnvironmentInfo object must be provided.

    The function connects the current container to the network of the container
    where the Docker DB is running.

    The names of created docker container, docker volume and docker network will be
    saved in the provided secret store. They will be used by the function that
    takes down the environment.

    The function saves the main AI-Lab configuration parameters, such as the DB and
    BucketFS connection parameters, in the secret store.
    """

    _remove_current_container_from_db_network(conf)

    if env_info is None:
        mem_size = f'{conf.get(AILabConfig.mem_size, "4")} GiB'
        disk_size = f'{conf.get(AILabConfig.disk_size, "10")} GiB'
        db_version = os.getenv(TEST_DB_VERSION_ENV_VAR, LATEST_DB_VERSION)
        accelerator = conf.get(AILabConfig.accelerator, Accelerator.none.value)

        docker_environment_variable: tuple[str, ...] = ()
        additional_db_parameter: tuple[str, ...] = ("-etlCheckCertsDefault=0",)
        itde_accelerator: tuple[str, ...] = ()
        if accelerator == Accelerator.nvidia.value:
            additional_db_parameter = additional_db_parameter + (
                "-enableAcceleratorDeviceDetection=1",
            )
            itde_accelerator = ("nvidia=all",)

        env_info, _ = api.spawn_test_environment(
            environment_name=ENVIRONMENT_NAME,
            nameserver=(NAME_SERVER_ADDRESS,),
            db_mem_size=mem_size,
            db_disk_size=disk_size,
            docker_db_image_version=db_version,
            docker_environment_variable=docker_environment_variable,
            accelerator=itde_accelerator,
            additional_db_parameter=additional_db_parameter,
            log_level=logging.getLevelName(logging.INFO),
        )

    db_info = env_info.database_info
    container_info = db_info.container_info

    assert container_info is not None
    _add_current_container_to_db_network(container_info.network_info.network_name)

    conf.save(AILabConfig.itde_container, container_info.container_name)
    if container_info.volume_name is not None:
        conf.save(AILabConfig.itde_volume, container_info.volume_name)
    conf.save(AILabConfig.itde_network, env_info.network_info.network_name)

    conf.save(AILabConfig.db_host_name, db_info.host)
    conf.save(AILabConfig.bfs_host_name, db_info.host)
    conf.save(AILabConfig.db_port, str(db_info.ports.database))
    conf.save(AILabConfig.bfs_port, str(db_info.ports.bucketfs))

    # Q. Can we draw any of the below constants from the ITDE configuration?
    conf.save(AILabConfig.db_user, "sys")
    conf.save(AILabConfig.db_password, "exasol")
    conf.save(AILabConfig.bfs_user, "w")
    conf.save(AILabConfig.bfs_password, "write")
    conf.save(AILabConfig.bfs_service, "bfsdefault")
    conf.save(AILabConfig.bfs_bucket, "default")
    conf.save(AILabConfig.db_encryption, "True")
    # The BucketFS encryption is turned off temporarily.
    conf.save(AILabConfig.bfs_encryption, "False")
    conf.save(AILabConfig.cert_vld, "False")


def _get_current_container(docker_client: docker.DockerClient):
    ip_addresses = _get_ipv4_addresses()
    return ContainerByIp(docker_client).find(ip_addresses)


def _add_current_container_to_db_network(network_name: str) -> None:
    with ContextDockerClient() as docker_client:
        container = _get_current_container(docker_client)
        if not container:
            return
        network = _get_docker_network(docker_client, network_name)
        if network and not _is_container_connected_to_network(container, network):
            network.connect(container.id)


def _is_container_connected_to_network(container, network) -> bool:
    network.reload()
    return container in network.containers


def _is_current_container_visible(network_name: str) -> bool:
    """
    For the Docker Edition returns True if the current (AI-Lab) container
    is connected to the network with the specified name, otherwise False.
    For other editions it always returns True.
    """
    with ContextDockerClient() as docker_client:
        container = _get_current_container(docker_client)
        if not container:
            # Not the Docker Edition
            return True
        network = _get_docker_network(docker_client, network_name)
        if not network:
            return False
        return _is_container_connected_to_network(container, network)


def _get_docker_network(
    docker_client: docker.DockerClient, network_name: str
) -> Network | None:
    networks = docker_client.networks.list(names=[network_name])
    if len(networks) == 1:
        network = networks[0]
        return network
    return None


def _remove_current_container_from_db_network(conf: Secrets):
    network_name = conf.get(AILabConfig.itde_network)
    if not network_name:
        return
    with ContextDockerClient() as docker_client:
        ip_addresses = _get_ipv4_addresses()
        container = ContainerByIp(docker_client).find(ip_addresses)
        if not container:
            return
        network = _get_docker_network(docker_client, network_name)
        if network and _is_container_connected_to_network(container, network):
            network.disconnect(container.id)


def _get_ipv4_addresses():
    ip_addresses = [
        ip.ip for ip in IPRetriever().ips() if ip.is_IPv4 and isinstance(ip.ip, str)
    ]
    return ip_addresses


def get_itde_status(conf: Secrets) -> ItdeContainerStatus:
    """
    Checks if the ITDE container exists and ready to be used. In the Docker Edition that
    means the ITDE is running and the AI-Lab container is connected to its network. In
    other editions it will just check that the ITDE is running.

    Returns the container status.

    The name of the container is taken from the provided secret store.
    If the name cannot be found there the function returns the status ABSENT.
    """

    # Try to get the names of the Docker-DB container and its network from the secret store.
    container_name = conf.get(AILabConfig.itde_container)
    network_name = conf.get(AILabConfig.itde_network)
    if not container_name or not network_name:
        return ItdeContainerStatus.ABSENT

    # Check the existence and the status of the container using the Docker API.
    with ContextDockerClient() as docker_client:
        if docker_client.containers.list(all=True, filters={"name": container_name}):
            container = docker_client.containers.get(container_name)
            if container.status != "running":
                return ItdeContainerStatus.STOPPED
            status = ItdeContainerStatus.RUNNING
            if _is_current_container_visible(network_name):
                status |= ItdeContainerStatus.VISIBLE
            return status
        return ItdeContainerStatus.ABSENT


def restart_itde(conf: Secrets) -> None:
    """
    Starts an existing ITDE container if it's not already running. In the Docker Edition
    connects the AI-Lab container to the Docker-DB network, unless it's already connected
    to it.

    For this function to work the container must exist. If it doesn't a RuntimeError will
    be raised. Use the get_itde_status function to check if the container exists.
    """

    status = get_itde_status(conf)

    if status is ItdeContainerStatus.ABSENT:
        raise RuntimeError("The Docker-DB container doesn't exist.")

    if ItdeContainerStatus.RUNNING not in status:  # type: ignore[operator]
        container_name = conf.get(AILabConfig.itde_container)
        with ContextDockerClient() as docker_client:
            container = docker_client.containers.get(container_name)
            container.start()

    if ItdeContainerStatus.VISIBLE not in status:  # type: ignore[operator]
        network_name = conf.get(AILabConfig.itde_network)
        if network_name:
            _add_current_container_to_db_network(network_name)


def take_itde_down(conf: Secrets, stop_db: bool = True) -> None:
    """
    Shuts down the ITDE.
    The names of the docker container, docker volume and docker network
    are taken from the provided secret store. If the names are not found
    there no action is taken.

    Stopping the DockerDB can be skipped by setting the stop_db parameter to False.
    This is a useful option in case the DockerDB has been provided externally.
    """
    _remove_current_container_from_db_network(conf)

    if stop_db:
        remove_container(conf)
        remove_volume(conf)
        remove_network(conf)

    conf.remove(AILabConfig.db_host_name)
    conf.remove(AILabConfig.bfs_host_name)
    conf.remove(AILabConfig.db_port)
    conf.remove(AILabConfig.bfs_port)
    conf.remove(AILabConfig.db_user)
    conf.remove(AILabConfig.db_password)
    conf.remove(AILabConfig.bfs_user)
    conf.remove(AILabConfig.bfs_password)
    conf.remove(AILabConfig.bfs_service)
    conf.remove(AILabConfig.bfs_bucket)
    conf.remove(AILabConfig.db_encryption)
    conf.remove(AILabConfig.bfs_encryption)
    conf.remove(AILabConfig.cert_vld)


def remove_network(conf):
    network_name = conf.get(AILabConfig.itde_network)
    if network_name:
        remove_docker_networks(iter([network_name]))
        conf.remove(AILabConfig.itde_network)


def remove_volume(conf):
    volume_name = conf.get(AILabConfig.itde_volume)
    if volume_name:
        remove_docker_volumes([volume_name])
        conf.remove(AILabConfig.itde_volume)


def remove_container(conf):
    container_name = conf.get(AILabConfig.itde_container)
    if container_name:
        remove_docker_container([container_name])
        conf.remove(AILabConfig.itde_container)



from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import (
    ScsOption,
    ScsSecretOption,
)

ONPREM_DB_OPTIONS = [
    ScsOption(
        "--db-host-name",
        metavar="HOST",
        type=str,
        help="Database connection host name",
        scs_key=CKey.db_host_name,
    ),
    ScsOption(
        "--db-port",
        metavar="PORT",
        type=int,
        default=8563,
        help="Database connection port",
        scs_key=CKey.db_port,
    ),
    ScsOption(
        "--db-username",
        metavar="USERNAME",
        type=str,
        help="Database user name",
        scs_key=CKey.db_user,
    ),
    ScsSecretOption(
        "--db-password",
        envvar="SCS_EXASOL_DB_PASSWORD",
        prompt="Exasol database password",
        scs_key=CKey.db_password,
    ),
    ScsOption(
        "--db-use-encryption/--no-db-use-encryption",
        type=bool,
        default=True,
        help="Whether to encrypt communication with the database or not",
        scs_key=CKey.db_encryption,
    ),
]



import random
import string
import traceback

import exasol.bucketfs as bfs

import exasol.nb_connector.cli.reporting as report
from exasol.nb_connector.cli.processing.option_set import ScsCliError
from exasol.nb_connector.connections import open_bucketfs_location
from exasol.nb_connector.secret_store import Secrets


def random_string(length: int = 10) -> str:
    return "".join(random.choice(string.ascii_uppercase) for _ in range(length))


def random_file_name(other_than: list[str]) -> str:
    result = None
    while result is None or result in other_than:
        infix = random_string()
        result = f"probe-{infix}.txt"
    return result


def files_in(bfsloc: bfs.path.PathLike) -> list[bfs.path.PathLike]:
    try:
        return list(bfsloc.iterdir())
    except FileNotFoundError:
        return []


def verify_bucketfs_access(scs: Secrets) -> None:
    bfs_root = open_bucketfs_location(scs)
    existing = [f.name for f in files_in(bfs_root)]
    file = bfs_root / random_file_name(other_than=existing)
    content = random_string(length=100)
    try:
        file.write(content.encode())
        actual = bfs.as_string(file.read())
        if actual != content:
            raise ScsCliError(f'{file} contains "{actual}" instead of "{content}".')
        report.success(
            "Access to the BucketFS of the configured database instance was successful."
        )
    except Exception as ex:
        stacktrace = traceback.format_exc()
        raise ScsCliError(f"Couldn't access the BucketFS: {stacktrace}") from ex
    finally:
        file.rm()



import ipywidgets as widgets

from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.ui.config.generic import get_config

DEFAULT_SCHEMA = "Default Schema"


def get_saas(conf: Secrets) -> widgets.Widget:
    """
    Creates a UI form for editing the Exasol SaaS database configuration.
    """

    inputs = [
        [
            (
                "Service URL",
                widgets.Text(value=conf.get(CKey.saas_url, "https://cloud.exasol.com")),
                CKey.saas_url,
            ),
            (
                "Account ID",
                widgets.Password(value=conf.get(CKey.saas_account_id)),
                CKey.saas_account_id,
            ),
            (
                "Database ID",
                widgets.Text(value=conf.get(CKey.saas_database_id)),
                CKey.saas_database_id,
            ),
            (
                "Database Name",
                widgets.Text(value=conf.get(CKey.saas_database_name)),
                CKey.saas_database_name,
            ),
            (
                "Personal Access Token",
                widgets.Password(value=conf.get(CKey.saas_token)),
                CKey.saas_token,
            ),
            (
                "Default Schema",
                widgets.Text(value=conf.get(CKey.db_schema, "AI_LAB")),
                CKey.db_schema,
            ),
        ],
        [
            (
                "Validate Certificate",
                widgets.Checkbox(
                    value=conf.get(CKey.cert_vld, "True") == "True", indent=False
                ),
                CKey.cert_vld,
            ),
            (
                "Trusted CA File/Dir",
                widgets.Text(value=conf.get(CKey.trusted_ca)),
                CKey.trusted_ca,
            ),
        ],
    ]

    group_names = ["SaaS DB Configuration", "TLS/SSL Configuration"]

    return get_config(conf, inputs, group_names)



from __future__ import annotations

from exasol.slc.models.compression_strategy import CompressionStrategy

from exasol.nb_connector.secret_store import Secrets
from exasol.nb_connector.slc.slc_error import SlcError


class SlcCompressionStrategy:
    def __init__(self, slc_name):
        self.slc_name = slc_name

    @property
    def key(self):
        return f"SLC_COMPRESSION_STRATEGY_{self.slc_name.upper()}"

    def save(self, secrets: Secrets, compression_strategy: CompressionStrategy) -> None:
        secrets.save(self.key, compression_strategy.value)

    def exists(self, secrets: Secrets) -> bool:
        return True if secrets.get(self.key) else False

    def verify(self, secrets: Secrets) -> CompressionStrategy:
        try:
            value = secrets[self.key]
            return CompressionStrategy(value)
        except AttributeError as ex:
            raise SlcError(
                "Secure Configuration Storage does not contain a"
                f" compression strategy for SLC {self.slc_name}."
            ) from ex



from exasol.nb_connector.ai_lab_config import AILabConfig as CKey
from exasol.nb_connector.cli.param_wrappers import (
    ScsOption,
    ScsSecretOption,
)

BUCKETFS_OPTIONS = [
    ScsOption(
        "--bucketfs-host",
        metavar="HOST",
        type=str,
        help="BucketFS host name",
        scs_key=CKey.bfs_host_name,
    ),
    ScsOption(
        "--bucketfs-host-internal",
        metavar="HOST",
        type=str,
        default="localhost",
        help="BucketFS internal host name",
        scs_key=CKey.bfs_internal_host_name,
    ),
    ScsOption(
        "--bucketfs-port",
        metavar="PORT",
        type=int,
        default=2580,
        help="BucketFS port",
        scs_key=CKey.bfs_port,
    ),
    ScsOption(
        "--bucketfs-port-internal",
        metavar="PORT",
        type=int,
        default=2580,
        help="BucketFS internal port",
        scs_key=CKey.bfs_internal_port,
    ),
    ScsOption(
        "--bucketfs-user",
        metavar="USERNAME",
        type=str,
        default="w",
        help="BucketFS user name",
        scs_key=CKey.bfs_user,
    ),
    ScsSecretOption(
        "--bucketfs-password",
        envvar="SCS_BUCKETFS_PASSWORD",
        prompt="BucketFS write password",
        scs_key=CKey.bfs_password,
    ),
    ScsOption(
        "--bucketfs-name",
        metavar="BFS_SERVICE",
        type=str,
        help='BucketFS service name, e.g. "bfsdefault"',
        scs_key=CKey.bfs_service,
    ),
    ScsOption(
        "--bucket",
        metavar="BUCKET",
        type=str,
        help='BucketFS bucket name, e.g. "default"',
        scs_key=CKey.bfs_bucket,
    ),
    ScsOption(
        "--bucketfs-use-encryption/--no-bucketfs-use-encryption",
        type=bool,
        default=True,
        help="Whether to encrypt communication with the BucketFS or not",
        scs_key=CKey.bfs_encryption,
    ),
]



"""
Github-related utility functions - check for latest release of
project, retrieval of artefacts, etc.
"""

import enum
import logging
import pathlib

import requests

_logger = logging.getLogger(__name__)


class Project(enum.Enum):
    """
    Names of github projects to be retrieved. Values have to
    match github project names.
    """

    CLOUD_STORAGE_EXTENSION = "cloud-storage-extension"
    KAFKA_CONNECTOR_EXTENSION = "kafka-connector-extension"


def get_latest_version_and_jar_url(project: Project) -> tuple[str, str]:
    """
    Retrieves the latest version of stable project release
    and url with jar file from the release.

    :param project: name of the project
    :return: tuple with version and url to retrieve the artefact.
    """
    req = requests.get(
        f"https://api.github.com/repos/exasol/{project.value}" f"/releases/latest",
        timeout=10,
    )
    if req.status_code != 200:
        raise RuntimeError(
            "Error sending request to the github, code: %d" % req.status_code
        )
    data = req.json()
    version = data.get("tag_name")
    if version is None:
        raise RuntimeError(
            f"The latest version of {project.value} " f"has no tag, something is wrong"
        )
    for asset in data.get("assets", []):
        name = asset["name"]
        if name.endswith(f"{version}.jar"):
            dl_url = asset["browser_download_url"]
            return version, dl_url
    raise RuntimeError("Could not find proper jar url for the latest release")


def retrieve_jar(
    project: Project,
    use_local_cache: bool = True,
    storage_path: pathlib.Path | None = None,
) -> pathlib.Path:
    """
    Returns latest jar file for the project, possibly using local cache.

    :param project: project to be used
    :param use_local_cache: should local cache be used or file always retrieved anew
    :param storage_path: path to be used for downloading.
        If None, current directory will be used.
    :return: path to the jar file on the local filesystem
    """
    version, jar_url = get_latest_version_and_jar_url(project)
    _, local_jar_name = jar_url.rsplit("/", maxsplit=1)
    local_jar_path = pathlib.Path(local_jar_name)
    if storage_path is not None:
        if not storage_path.exists():
            raise ValueError(f"Local storage path doesn't exist: {storage_path}")
        local_jar_path = storage_path / local_jar_path

    if use_local_cache and local_jar_path.exists():
        _logger.info(
            "Jar for version %s already exists in %s, skip downloading",
            version,
            local_jar_path,
        )
    else:
        _logger.info("Fetching jar for version %s from %s...", version, jar_url)
        req = requests.get(jar_url, stream=True, timeout=10)
        try:
            count_bytes = local_jar_path.write_bytes(req.content)
            _logger.info("Saved %d bytes in %s", count_bytes, local_jar_path)
        finally:
            req.close()
    return local_jar_path
